"""
This file provides utility functions for symmetry adapted model.
"""
import sympy as sp
import numpy as np
import multiprocessing
from joblib import Parallel, delayed

from gcoreutils.nsarray import NSArray

from multipie.tag.tag_list import TagList
from multipie.multipole.util.structure_samb_util import orthogonalize_fk, decompose_fk


# number of cpu cores
_cpu_num = multiprocessing.cpu_count()


# ==================================================
def create_atomic_samb_set(pg, atomic_braket, spinful, parallel=True):
    """
    create atomic multipole basis set for each atomic subspaces.

    Args:
        pg (PointGroup): point group.
        atomic_braket (dict): { "M_#": (bra_list, ket_list) }.
        spinful (bool): spinful ?
        parallel (bool, optional): use parallel code ?

    Returns:
        tuple: information of atomic multipoles, (atomic_info, atomic_data).

    Note:
        atomic_info = { "M_#": ["amp_#"] }
        atomic_data = { "amp_#" : (TagMultipole, NSArray(matrix)) }
    """
    M_num = len(atomic_braket)
    n_jobs = max(abs(min(M_num + 1, _cpu_num - 2)), 1) if parallel else 1

    def proc(i, M_i):
        bra_list, ket_list = atomic_braket[M_i]
        atomic_samb = pg.atomic_samb(bra_list, ket_list, spinful)
        return i, M_i, atomic_samb

    M_i_list = list(atomic_braket.keys())
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc)(i, M_i) for i, M_i in enumerate(M_i_list))
    res.sort(key=lambda x: x[0])

    atomic_info = {}
    atomic_data = {}
    i = 1
    for _, M_i, atomic_samb in res:
        lst = []
        for tag, m in atomic_samb.items():
            lst.append(f"amp_{i:03d}")
            atomic_data[f"amp_{i:03d}"] = (tag, m)
            i += 1
        atomic_info[M_i] = lst

    return atomic_info, atomic_data


# ==================================================
def create_site_cluster_samb_set(g, rep_site_dict, parallel=True):
    """
    create site-cluster multipole basis set for each S_#.

    Args:
        g (PointGroup/SpaceGroup): point/space group.
        rep_site_dict (dict): { cluster_tag: position }.
        parallel (bool, optional): use parallel code ?

    Returns:
        dict: { "S_#": (TagMultipole, NSArray(vector)) }
    """
    S_num = len(rep_site_dict)
    n_jobs = max(abs(min(S_num + 1, _cpu_num - 2)), 1) if parallel else 1

    def proc(i, S_i):
        rep_site = rep_site_dict[S_i]
        site_cluster_samb = g.site_cluster_samb(rep_site)[0]
        return i, S_i, site_cluster_samb

    S_i_list = list(rep_site_dict.keys())
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc)(i, S_i) for i, S_i in enumerate(S_i_list))
    res.sort(key=lambda x: x[0])

    scm_set = {S_i: site_cluster_samb for _, S_i, site_cluster_samb in res}

    return scm_set


# ==================================================
def create_bond_cluster_samb_set(g, rep_bond_dict, parallel=True):
    """
    create bond-cluster multipole basis set for each B_#.

    Args:
        g (PointGroup/SpaceGroup): point/space group.
        rep_bond_dict (dict): { cluster_tag: vector@center }.
        parallel (bool, optional): use parallel code ?

    Returns:
        dict: { "B_#": (TagMultipole, NSArray(vector)) }
    """
    B_num = len(rep_bond_dict)
    n_jobs = max(abs(min(B_num + 1, _cpu_num - 2)), 1) if parallel else 1

    def proc(i, B_i):
        rep_bond = rep_bond_dict[B_i]
        bond_cluster_samb = g.bond_cluster_samb(rep_bond)[0]
        return i, B_i, bond_cluster_samb

    B_i_list = list(rep_bond_dict.keys())
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc)(i, B_i) for i, B_i in enumerate(B_i_list))
    res.sort(key=lambda x: x[0])

    bcm_set = {B_i: bond_cluster_samb for _, B_i, bond_cluster_samb in res}

    return bcm_set


# ==================================================
def create_cluster_samb_set(g, rep_site_dict, rep_bond_dict, parallel=True):
    """
    create site-cluster multipole basis set for each S_#.

    Args:
        g (PointGroup/SpaceGroup): point/space group.
        rep_site_dict (dict): { cluster_tag: position }.

        rep_bond_dict (dict): { cluster_tag: vector@center }.
        parallel (bool, optional): use parallel code ?

    Returns:
        tuple: information of site/bond-cluster multipoles,
                (site_cluster_info, site_cluster_data, bond_cluster_info, bond_cluster_data).

    Note:
        site_cluster_info = { "S_#": ["smp_#"] }
        site_cluster_data = { "smp_#" : (TagMultipole, NSArray(vector)) ] }
        bond_cluster_info = { "B_#": ["bmp_#"] }
        bond_cluster_data = { "bmp_#" : (TagMultipole, NSArray(vector)) ] }
    """
    scm_set = create_site_cluster_samb_set(g, rep_site_dict, parallel)
    bcm_set = create_bond_cluster_samb_set(g, rep_bond_dict, parallel)

    site_cluster_info = {}
    site_cluster_data = {}
    i = 1
    for S_i, site_cluster_samb in scm_set.items():
        lst = []
        for tag, v in site_cluster_samb.items():
            lst.append(f"smp_{i:03d}")
            site_cluster_data[f"smp_{i:03d}"] = (tag, v)
            i += 1
        site_cluster_info[S_i] = lst

    bond_cluster_info = {}
    bond_cluster_data = {}
    for B_i, bond_cluster_samb in bcm_set.items():
        lst = []
        for tag, v in bond_cluster_samb.items():
            lst.append(f"bmp_{i:03d}")
            bond_cluster_data[f"bmp_{i:03d}"] = (tag, v)
            i += 1
        bond_cluster_info[B_i] = lst

    return site_cluster_info, site_cluster_data, bond_cluster_info, bond_cluster_data


# ==================================================
def uniform_samb(sbc_samb, braket_indexes, dim):
    """
    create uniform multipole basis set.

    Args:
        sbc_samb (list): site/bond-cluster multipole basis set, [(TagMultipole, NSArray(vector))].
        braket_indexes (list): [(bra_site_no, ket_site_no)]
        dim (int): dimension of matrix.

    Returns:
        dict: {TagMultipole: NSArray(matrix)}.
    """
    u_samb = {}
    for tag, v in sbc_samb:
        U = sp.Matrix.zeros(dim, dim)
        for vi, (bra_idx, ket_idx) in zip(v, braket_indexes):
            U[bra_idx, ket_idx] += vi

        if not U.is_diagonal():
            U = (U + U.adjoint()) / sp.sqrt(2)

        U = NSArray(str(U.tolist()), style="matrix", fmt="sympy").simplify()
        if not np.all(U == 0):
            u_samb[tag] = U

    return u_samb


# ==================================================
def create_uniform_samb_set(cluster_samb_set, braket_indexes_dict, dim, parallel=True):
    """
    create uniform multipole basis set.

    Args:
        cluster_samb_set (dict): { "S_#"/"B_#": [(TagMultipole, NSArray)] }.
        braket_indexes_dict (dict): { cluster_tag: [(bra_site_no, ket_site_no)] }.
        dim (int): dimension of matrix.
        parallel (bool, optional): use parallel code ?

    Returns:
        tuple: information of uniform multipoles, uniform_info, uniform_data.

    Note:
        uniform_info = { "S_#"/"B_#": ["ump_#"] }
        uniform_data = { "ump_#" : (TagMultipole, NSArray(matrix)) }
    """
    SB_num = len(cluster_samb_set)
    n_jobs = max(abs(min(SB_num + 1, _cpu_num - 2)), 1) if parallel else 1

    def proc(i, SB_i, braket_indexes):
        sbc_samb = cluster_samb_set[SB_i]
        is_diagonal = all([bra_idx == ket_idx for (bra_idx, ket_idx) in braket_indexes])
        head_list = ["Q"] if is_diagonal else ["Q", "T"]
        sbc_samb = [(tag, v) for tag, v in sbc_samb if tag.head in head_list]
        if len(sbc_samb) == 0:
            return i, SB_i, {}

        u_samb = uniform_samb(sbc_samb, braket_indexes, dim)
        return i, SB_i, u_samb

    res = Parallel(n_jobs=n_jobs, verbose=0)(
        delayed(proc)(i, SB_i, braket_indexes) for i, (SB_i, braket_indexes) in enumerate(braket_indexes_dict.items())
    )
    res.sort(key=lambda x: x[0])

    dic = {SB_i: u_samb for _, SB_i, u_samb in res}

    # orthogonalization
    def proc_ortho(i, head):
        tags = [(SB_i, tag) for SB_i, u_samb in dic.items() for tag in u_samb.keys() if tag.head == head]
        mats = [dic[SB_i][tag] for (SB_i, tag) in tags]
        if len(mats) == 0:
            return i, {}

        mats = NSArray(mats, style="matrix", fmt="sympy", real=False)
        mats, idx = NSArray.orthogonalize(mats)
        tags, mats = [tags[j] for j in idx], [mats[j] for j in idx]
        d = {}
        for (SB_i, tag), U in zip(tags, mats):
            if sp.Matrix(U).is_diagonal():
                tag = tag.replace(m_type="s")
            else:
                tag = tag.replace(m_type="u")
            d[(SB_i, tag)] = U

        um_orthogonalized = {
            (SB_i, tag): NSArray(mat, style="matrix", fmt="sympy", real=False) for (SB_i, tag), mat in d.items()
        }
        return i, um_orthogonalized

    head_list = ["Q", "T"]
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc_ortho)(i, head) for i, head in enumerate(head_list))
    res.sort(key=lambda x: x[0])

    uniform_info = {}
    uniform_data = {}
    i = 1
    for _, u_samb in res:
        for (SB_i, tag), v in u_samb.items():
            uniform_data[f"ump_{i:03d}"] = (tag, v)
            if SB_i in uniform_info:
                uniform_info[SB_i] += [f"ump_{i:03d}"]
            else:
                uniform_info[SB_i] = [f"ump_{i:03d}"]
            i += 1

    return uniform_info, uniform_data


# ==================================================
def structure_samb(bc_samb, bond_list, bond):
    """
    create structure multipole basis set.

    Args:
        bc_samb (list): bond-cluster multipole basis set, [(TagMultipole, NSArray(vector))].
        bond_list (list) : ["bond_#"].
        bond (dict, optional): { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }.

    Returns:
        dict: {TagMultipole: Symbol}.
    """
    n_sgn_list = []
    for bond_n in bond_list:
        vec = bond[bond_n][3]
        if "bond" in vec:
            n = int(vec.split("_")[1])
        else:
            n = int(bond_n.split("_")[1])

        if vec[0] == "-":
            n_sgn_list.append((n, +1))
        else:
            n_sgn_list.append((n, -1))

    fk_list = []
    tags = []
    for tag, v in bc_samb:
        fk = sp.S(0)
        for vn, (n, sgn) in zip(v, n_sgn_list):
            cn = sp.Symbol(f"c{n:03d}", real=True)
            sn = sp.Symbol(f"s{n:03d}", real=True)
            v = vn * (cn + sgn * sp.I * sn)
            fk += v + sp.conjugate(v)
        tags.append(tag)
        fk_list.append(sp.expand(fk))

    # orthogonalization
    fk_list, idx = orthogonalize_fk(fk_list)
    fk_list = [fk_list[i] for i in idx]
    tags = [tags[i] for i in idx]

    k_samb = {tag.replace(m_type="k"): fk for tag, fk in zip(tags, fk_list)}

    return k_samb


# ==================================================
def create_structure_samb_set(bc_samb_set, cluster_bond, bond, parallel=True):
    """
    create structure multipole basis set.

    Args:
        bc_samb_set (dict): { "B_#": [(TagMultipole, NSArray)] }.
        cluster_bond (dict, optional) : { cluster_tag: bond_list }.
        bond (dict, optional): { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }.
        parallel (bool, optional): use parallel code ?

    Returns:
        tuple: information of structure multipoles, structure_info, structure_data.

    Note:
        structure_info = { "B_#": ["kmp_#"] }
        structure_data = { "kmp_#" : (TagMultipole, Symbol) }
    """
    B_num = len(bc_samb_set)
    n_jobs = max(abs(min(B_num + 1, _cpu_num - 2)), 1) if parallel else 1

    def proc(i, B_i):
        bc_samb = bc_samb_set[B_i]
        bond_list = cluster_bond[B_i]
        k_samb = structure_samb(bc_samb, bond_list, bond)
        return (i, B_i, k_samb)

    B_i_list = list(cluster_bond.keys())
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc)(i, B_i) for i, B_i in enumerate(B_i_list))
    res.sort(key=lambda x: x[0])

    structure_info = {}
    structure_data = {}
    i = 1
    for _, B_i, k_samb in res:
        lst = []
        for tag, fk in k_samb.items():
            lst.append(f"kmp_{i:03d}")
            structure_data[f"kmp_{i:03d}"] = (tag, sp.expand(fk))
            i += 1
        structure_info[B_i] = lst

    return structure_info, structure_data


# ==================================================
def _check_complete_relation(x_tag_dict, y_tag_dict, z_info, z_data):
    """
    check complete relation of combined multipole basis set.

    Args:
        x_tag_dict (dict): multipole/harmonics tag dict, {M_#: TagList}.
        y_tag_dict (dict): multipole/harmonics tag dict, {S_#/B_#: TagList}.
        z_info (dict): { ("S_#"/"B_#", "M_#"): ["z_#"] }.
        z_data (dict): { "z_#" : [(coefficient, "amp_#", "smp_#"/"bmp_#"/"ump_#")] }.
    """
    for M_i, SB_i in z_info.keys():
        tag1_list = x_tag_dict[M_i]
        tag2_list = y_tag_dict[SB_i]
        Z_lst = [z_data[z_i] for z_i in z_info[(M_i, SB_i)]]
        n, n1, n2 = len(Z_lst), len(tag1_list), len(tag2_list)
        if n != n1 * n2:
            s = f"(M_i, SB_i) = {(M_i, SB_i)} \n"
            s += f"tag1_list = {[str(tag1) for tag1 in tag1_list]} \n"
            s += f"tag2_list = {[str(tag2) for tag2 in tag2_list]} \n"
            s += "tag2_list \n"
            for tag, lst in Z_lst:
                s += f"{tag} = {[(c, str(tag1), str(tag2)) for c, tag1, tag2 in lst]} \n"

            s += f"# of X = n1 = {n1} \n"
            s += f"# of Y = n2 = {n2} \n"
            s += f"# of Z = n = {n} \n"
            s += f"n1*n2 = n = {n} \n"
            s += f"n1*n2 = {n1*n2} must equal to n = {n}."
            raise Exception(s)


# ==================================================
def create_z_samb_set(
    g, x_tag_dict, y_tag_dict, M_SB_list, atomic_braket, toroidal_priority=False, parallel=True, **kwargs
):
    """
    create combined multipole basis set.

    Args:
        g (PointGroup/SpaceGroup): point/space group.
        x_tag_dict (dict): atomic multipole tag dict, {(M_#, TagMultipole): amp_i}.
        y_tag_dict (dict): site/bond-cluster tag dict, {(S_#/B_#, TagMultipole): smp_i/bmp_i/ump_i}.
        M_SB_list (list): [("M_#", "S_#"/"B_#")].
        atomic_braket (dict): { matrix_tag : (bra_list, ket_list) }.
        toroidal_priority (bool): create toroidal multipoles (G,T) in priority? else prioritize conventional multipoles (Q,M).
        parallel (bool, optional): use parallel code.
        kwargs (dict, optional): select conditions for multipoles,
                                 keywords in ["head", "rank", "irrep", "mul", "comp", "s", "k"].
    Returns:
        tuple: information of combined multipoles, z_info, z_data.

    Note:
        z_info = { ("S_#"/"B_#", "M_#"): ["z_#"] }
        z_data = { "z_#" : [(coefficient, "amp_#", "smp_#"/"bmp_#"/"ump_#")] }
    """
    E_am_dict = {}
    for M_i, SB_i in M_SB_list:
        o1, o2 = atomic_braket[M_i]
        if SB_i[0] == "B" and o1 != o2:
            E_am_dict[(M_i, SB_i)] = True
        else:
            E_am_dict[(M_i, SB_i)] = False

    M_SB_num = len(M_SB_list)
    n_jobs = max(abs(min(M_SB_num + 1, _cpu_num - 2)), 1) if parallel else 1

    irrep_list = kwargs.pop("irrep", None)
    if type(irrep_list) == str:
        irrep_list = [irrep_list]
    elif irrep_list is None:
        irrep_list = g.character.irrep_list if g.tag.pg == "" else g.pg.character.irrep_list
        irrep_list = [str(irrep) for irrep in irrep_list]

    irrep_M_SB_list = [(irrep, M_i, SB_i) for irrep in irrep_list for (M_i, SB_i) in M_SB_list]

    def proc(i, irrep, M_i, SB_i):
        x_tag_list = TagList([tag for (M_i_, tag) in x_tag_dict.keys() if M_i_ == M_i])
        if E_am_dict[(M_i, SB_i)]:
            x_tag_list = TagList([tag for tag in x_tag_list if tag.head in ("Q", "G")])
        y_tag_list = TagList([tag for (SB_i_, tag) in y_tag_dict.keys() if SB_i_ == SB_i])
        z_samb = g.z_samb(x_tag_list, y_tag_list, toroidal_priority, irrep=irrep, **kwargs)
        return i, irrep, M_i, SB_i, z_samb

    res = Parallel(n_jobs=n_jobs, verbose=0)(
        delayed(proc)(i, irrep, M_i, SB_i) for i, (irrep, M_i, SB_i) in enumerate(irrep_M_SB_list)
    )
    res.sort(key=lambda x: x[0])

    z_info = {}
    z_data = {}
    i = 1
    for _, irrep, M_i, SB_i, z_samb in res:
        lst = []
        for tags, c_xtag_ytag_list in z_samb.items():
            tag, _, _ = tags
            lst.append(f"z_{i:03d}")
            c_amp_cmp_list = [
                (c, x_tag_dict[(M_i, xtag)], y_tag_dict[(SB_i, ytag)]) for c, xtag, ytag in c_xtag_ytag_list
            ]
            z_data[f"z_{i:03d}"] = (tag, c_amp_cmp_list)
            i += 1
        if lst != []:
            z_info[(irrep, M_i, SB_i)] = lst

    # check basis (# of X)×(# of Y) = # of Z.
    if len(kwargs) == 0:
        _check_complete_relation(x_tag_dict, y_tag_dict, z_info, z_data)

    return z_info, z_data


# ==================================================
def _fourier_transform_bond_cluster_samb(bc_samb, u_samb_set, k_samb_set, braket_indexes, bond_list, bond, dim):
    """
    fourier series of bond-cluster multipoles.

    Yb_{a} => Σ_{bc} C_{abc} U_{b} fk_{c}

    Args:
        bc_samb (list): bond-cluster multipole basis set, [("bmp_#", NSArray(vector))].
        u_samb_set (list): uniform multipole basis set, [("ump_#", NSArray(matrix))].
        k_samb_set (list): structure multipole basis set, [("kmp_#", Symbol)].
        braket_indexes (list): [(bra_site_no, ket_site_no)].
        bond_list (list) : ["bond_#"].
        bond (dict, optional): { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }.
        dim (int): dimension of site-space matrix.

    Returns:
        dict: fourier series of bond-cluster multipoles, { "bmp_#" : [(coefficient, "ump_#", "kmp_#")] }.
    """

    def _check_fourier_transform(coeff_list):
        v = sp.Matrix(coeff_list)
        v = sp.adjoint(v).dot(v)
        norm = sp.sqrt(sp.simplify(v))

        if norm != 1:
            raise Exception(f"invalid fourier expansion coefficient = {v}, norm = {norm}.")

    n_sgn_list = []
    for bond_n in bond_list:
        vec = bond[bond_n][3]
        if "bond" in vec:
            n = int(vec.split("_")[1])
        else:
            n = int(bond_n.split("_")[1])

        if vec[0] == "-":
            n_sgn_list.append((n, +1))
        else:
            n_sgn_list.append((n, -1))

    Mfk_set = {}
    for bmp_i, v in bc_samb:
        Mk = sp.Matrix.zeros(dim, dim)
        for vi, (bra_idx, ket_idx), (n, sgn) in zip(v, braket_indexes, n_sgn_list):
            cn = sp.Symbol(f"c{n:03d}", real=True)
            sn = sp.Symbol(f"s{n:03d}", real=True)
            Mk[bra_idx, ket_idx] += vi * (cn + sgn * sp.I * sn)

        Mk = (Mk + Mk.adjoint()) / sp.sqrt(2)
        Mfk_set[bmp_i] = sp.simplify(Mk)

    bck_samb = {}
    for bmp_i, Mk in Mfk_set.items():
        coeff_ump_smp_list = []
        for ump_i, m in u_samb_set:
            fk = (sp.Matrix(m).adjoint() * Mk).trace()
            fk = sp.simplify(fk)
            coeffs = decompose_fk(fk, k_samb_set)
            for kmp_i, c in coeffs.items():
                coeff_ump_smp_list.append((c, ump_i, kmp_i))

        # check normalization
        coeff_list = [c for c, _, _ in coeff_ump_smp_list]
        _check_fourier_transform(coeff_list)

        bck_samb[bmp_i] = coeff_ump_smp_list

    return bck_samb


# ==================================================
def fourier_transform_bond_cluster_samb(bc_samb_set, u_samb_set, k_samb_set, cluster_bond, bond, dim, parallel=True):
    """
    fourier series of bond-cluster multipoles.

    Yb_{a} => Σ_{bc} C_{abc} U_{b} fk_{c}

    Args:
        bc_samb_set (dict): { "B_#": [("bmp_#", NSArray(vector))] }.
        u_samb_set (list): uniform multipole basis set, [("ump_#", NSArray(matrix))].
        k_samb_set (list): structure multipole basis set, [("kmp_#", Symbol)].
        cluster_bond (dict) : { cluster_tag: bond_list }.
        bond (dict): { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }.
        dim (int): dimension of matrix.
        parallel (bool, optional): use parallel code ?

    Returns:
        dict: { "bmp_#" : [(coefficient, "ump_#", "kmp_#")] }.
    """
    B_num = len(bc_samb_set)
    n_jobs = max(abs(min(B_num + 1, _cpu_num - 2)), 1) if parallel else 1

    braket_indexes_dict = {
        B_i: [bond[bond_tag][2] for bond_tag in bond_list] for B_i, bond_list in cluster_bond.items()
    }

    def proc(i, B_i):
        bc_samb = bc_samb_set[B_i]
        braket_indexes = braket_indexes_dict[B_i]
        bond_list = cluster_bond[B_i]
        bck_samb = _fourier_transform_bond_cluster_samb(
            bc_samb, u_samb_set, k_samb_set, braket_indexes, bond_list, bond, dim
        )
        return i, bck_samb

    B_i_list = list(cluster_bond.keys())
    res = Parallel(n_jobs=n_jobs, verbose=0)(delayed(proc)(i, B_i) for i, B_i in enumerate(B_i_list))
    res.sort(key=lambda x: x[0])

    bond_cluster_k_data = {}
    for _, bck_samb in res:
        bond_cluster_k_data |= bck_samb

    return bond_cluster_k_data


# ==================================================
def create_zk_samb_set(z_data, bc_samb_set, u_samb_set, k_samb_set, cluster_bond, bond, dim, parallel=True):
    """
    create k-space representation of combined multipole basis set.

    Zk = Σ_{abc} C_{abc} X_{a} U_{b} fk_{c}

    Args:
        z_data (dict): { "z_#": (TagMultipole, [(coeff, "amp_#", "smp_#"/"bmp_#")]) }.
        bc_samb_set (dict): { "B_#": [("bmp_#", NSArray(vector))] }.
        u_samb_set (list): uniform multipole basis set, [("ump_#", NSArray(matrix))].
        k_samb_set (list): structure multipole basis set, [("kmp_#", Symbol)].
        cluster_bond (dict) : { cluster_tag: bond_list }.
        bond (dict): { bond_tag: (vector@center, SO, (bra_site_no, ket_site_no), vector, tail;head) }.
        dim (int): dimension of matrix.
        parallel (bool, optional): use parallel code ?

    Returns:
        dict: { "z_#": (TagMultipole, [ (coeff, "amp_#", "ump_#", "kmp_#") ]) }.
    """
    bond_cluster_k_data = fourier_transform_bond_cluster_samb(
        bc_samb_set, u_samb_set, k_samb_set, cluster_bond, bond, dim, parallel
    )

    Zk_data = {}
    for z_i, (tag, lst) in z_data.items():
        lst_k = []
        for c1, amp_i, cmp_i in lst:
            if cmp_i[0] == "b":
                for c2, ump_i, kmp_i in bond_cluster_k_data[cmp_i]:
                    c = sp.simplify(sp.sympify(c1) * sp.sympify(c2))
                    lst_k.append((c, amp_i, ump_i, kmp_i))
            else:
                ump_i = cmp_i.replace("s", "u")
                lst_k.append((c1, amp_i, ump_i))

        Zk_data[z_i] = (tag, lst_k)

    return Zk_data


# ==================================================
def redefine_index(info, data, z_data, molecule=False, init_idx=1):
    """
    Redefine the serial number of multipoles.

    Args:
        info (dict): { "S_#"/"B_#"/"M_#": ["x_#"] }, x = amp/smp/bmp/ump/kmp.
        data (dict): { "x_#" : definition (NSArray/list) }.
        z_data (dict): { "z_#" : (TagMultipole, [(coefficient, "amp_#", "cmp_#")]) }/{ "z_#": (TagMultipole, [ (coeff, "amp_#", "ump_#", "kmp_#") ]) }.
        molecule (bool, optional): molecule ?
        init_idx (int): initial number of index.

    Returns:
        tuple: redefined information of multipoles, (info, data, z_data).
    """
    x = list(info.values())[0][0].split("_")[0]
    if molecule and x in ("smp", "bmp"):
        sb = "s" if x == "smp" else "b"
        x_i_list = [v[2].replace("u", sb) for lst in z_data.values() for v in lst[1] if v[2].split("_")[0] == "ump"]
    else:
        x_i_list = [x_i for lst in z_data.values() for v in lst[1] for x_i in v[1:] if x_i.split("_")[0] == x]
    x_i_list = sorted(list(set(x_i_list)))

    old_new_dict = {}
    new_info = {}
    new_data = {}
    i = init_idx
    for k, lst in info.items():
        new_lst = []
        for x_i_old in lst:
            if x_i_old in x_i_list:
                x_i_new = f"{x}_{i:03d}"
                old_new_dict[x_i_old] = x_i_new
                new_lst.append(x_i_new)
                new_data[x_i_new] = data[x_i_old]
                i += 1
        if len(new_lst) > 0:
            new_info[k] = new_lst

    new_z_data = {}
    for k, lst in z_data.items():
        tag = lst[0]
        new_lst = []
        for v in lst[1]:
            c = v[0]
            new_v = []
            for x_i_old in v[1:]:
                if x_i_old in old_new_dict:
                    x_i_new = old_new_dict[x_i_old]
                else:
                    x_i_new = x_i_old
                new_v.append(x_i_new)
            new_lst.append((c, *new_v))
        new_z_data[k] = (tag, new_lst)

    return new_info, new_data, new_z_data



"""
TagSymmetryOperation manages tag of symmetry operation.
"""
import sympy as sp
from dataclasses import dataclass
from multipie.tag.tag import Tag


# ==================================================
@dataclass(frozen=True, order=True)
class TagSymmetryOperation(Tag):
    """
    tag of symmetry operation.
    """

    mirror: bool = False  # mirror ?
    inversion: bool = False  # inversion ?
    n: int = 1  # rotation.
    axis: str = "[0,0,1]"  # rotation axis.
    t: str = "[0,0,0]"  # partial translation.

    t0: bool = True  # zero translation ?
    pg: bool = True  # point group ?

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        """
        format of symmetry-operation string (translation is omittable).
            inversion(-) n-fold rotation(2,3+/-,4+/-,6+/-)mirror(m) axis[n1n2n3] :translation[t1,t2,t3].
        """
        s = tag
        s = s.replace("[", "").replace("]", "")  # remove []

        st = s.split(":")
        s = st[0]

        inversion = False
        mirror = False

        pos = 0
        c = s[pos]
        if c == "-":
            inversion = True
            pos += 1
        elif c == "m":
            mirror = True
            pos += 1

        n = 1
        if not mirror:
            n = int(s[pos])
            pos += 1

        if n == 3 or n == 4 or n == 6:
            d = s[pos]
            pos += 1
            if d == "-":
                n = -n

        axis = "[0,0,1]"
        if n != 1 or mirror:
            axis = TagSymmetryOperation._parse_axis(s[pos:])

        # partial translation.
        if len(st) > 1:
            pg = False
            t = st[1].split(",")
        else:
            pg = True
            t = ["0"] * 3
        t0 = t == ["0", "0", "0"]
        t = f"[{t[0]},{t[1]},{t[2]}]"

        d = {"mirror": mirror, "inversion": inversion, "n": n, "axis": axis, "t": t, "t0": t0, "pg": pg}
        return d

    # ==================================================
    def __str__(self):
        mirror = self.mirror
        inversion = self.inversion
        n = self.n
        axis = self.axis
        t = self.t
        pg = self.pg

        s = ""
        if mirror:
            s += "m"
        else:
            if inversion:
                s += "-"
            s += str(abs(n))
        if abs(n) > 2:
            s += "+" if n > 0 else "-"
        if n != 1 or mirror:
            s += axis.replace(",", "")
        if not pg:
            s += ":" + t

        return s

    # ==================================================
    def latex(self):
        astr = self.axis.replace("[", "").replace("]", "").replace(",", "")
        s = ""
        if not self.pg:
            s += "\\{"
        if self.mirror:
            s += r"{\rm m}_{" + astr + "}"
        else:
            if self.inversion:
                s += "-"
            an = abs(self.n)
            s += str(an)
            us = False
            if an > 2:
                s += "^{+}" if self.n > 0 else "^{-}"
                us = True
            if an != 1:
                if us:
                    s += "_{\\,\\," + astr + "}"
                else:
                    s += "{}_{" + astr + "}"

        if not self.pg:
            s += "|"
            if self.t0:
                s += "0"
            else:
                t = self.t.strip("[]").split(",")
                t = [sp.latex(sp.S(i)) for i in t]
                s += " ".join(t)
            s += "\\}"

        return s

    # ==================================================
    def is_point_group(self):
        return self.pg

    # ==================================================
    def is_no_translation(self):
        return self.t0

    # ==================================================
    @classmethod
    def _parse_axis(cls, tag):
        """
        parse axis string.

        Args:
            tag (str): axis string (without [],comma,space).

        Returns:
            str: axis string as [x,y,z].
        """
        s = tag.replace("[", "").replace("]", "").replace(",", "")

        pos = 0
        axis = [0, 0, 1]
        for i in range(3):
            if s[pos] == "-":
                axis[i] = int(s[pos : pos + 2])
                pos += 2
            else:
                axis[i] = int(s[pos])
                pos += 1
        return f"[{axis[0]},{axis[1]},{axis[2]}]"



"""
This file provides utility functions for calculation of atomic multipole basis set.
"""
import numpy as np

from gcoreutils.nsarray import NSArray

from multipie.multipole.util.atomic_orbital_util import rank, basis_type
from multipie.multipole.util.spin_orbital_basis import _standard_basis
from multipie.multipole.util.multipole_util import matrix_sum
from multipie.tag.tag_list import TagList
from multipie.tag.tag_multipole import TagMultipole
from multipie.const import __def_dict__


# ==================================================
def _extract_subspace(am_data, bra_list, ket_list, spinful, crystal):
    """
    extract subspace for given bra_list and ket_list.
    32✖️32 (spinful) or 16✖️16 (spinless) matrix Xlmsk => Xlmsk[bra_list, ket_list].

    Args:
        am_data (dict): data of atomic multipoles, { TagMultipole: Matrix (32✖️32/16✖️16)}.
        bra_list ([str]): orbital list of bra.
        ket_list ([str]): orbital list of ket.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        dict: atomic multipoles (subspace).
    """
    bra_rank_lst = [rank(o, spinful, crystal) for o in bra_list]
    ket_rank_lst = [rank(o, spinful, crystal) for o in ket_list]
    max_rank = max(bra_rank_lst) + max(ket_rank_lst)
    b_type = basis_type(bra_list, crystal)

    orbitals_dataset = _standard_basis[spinful][b_type]
    bra_idx_list = [orbitals_dataset.index(o) for o in bra_list]
    ket_idx_list = [orbitals_dataset.index(o) for o in ket_list]

    am_extracted = {}
    for tag, Xlm in am_data.items():
        if tag.rank > (max_rank + 1):
            continue

        Xlm_extracted = Xlm[bra_idx_list, :][:, ket_idx_list]

        if not np.all(Xlm_extracted == 0):
            am_extracted[TagMultipole(str(tag))] = Xlm_extracted

    return am_extracted


# ==================================================
def _unitary_transform(am_lm, hs):
    """
    unitary transform the atomic multipoles.
    XlΓnγ = Σ_{lm} U_{lm, lΓnγ} Xlm

    Args:
        am_lm (dict): atomic multipoles, { TagMultipole: Xlm }.
        hs (HarmonicsPGRSet): a set of point-group harmonics (real version).

    Returns:
        dict: transformed atomic multipoles (XlΓnγ).
    """
    shape = list(am_lm.values())[0].shape
    zm = NSArray.zeros(shape=shape, style="matrix", fmt="sympy")

    lXsk_list = [(t.l, t.head, t.s, t.k) for t in am_lm.keys()]
    lXsk_list = sorted(set(lXsk_list), key=lXsk_list.index)

    am_transformed = {}
    for l, X, s, k in lXsk_list:
        Xlm_list = []
        for m in reversed(range(-l, l + 1)):
            tag = TagMultipole.create_spherical(head=X, rank=l, mul=0, comp=m, s=s, k=k)
            Xlm_list.append(am_lm[tag]) if tag in am_lm else Xlm_list.append(zm)

        U = {h.tag: h.u_matrix() for h in hs.select(rank=l) if h.tag.i_type == __def_dict__["head_i"][X]}
        am_lg = {}
        for utag, u in U.items():
            Xlg = matrix_sum([Xlm_list[j] * u[j] for j in range(2 * l + 1)])
            if not np.all(Xlg == 0):
                tag = utag.replace(head=X, s=s, k=k, m_type="a")
                am_lg[tag] = Xlg.simplify()

        am_transformed.update(am_lg)

    return am_transformed


# ==================================================
def _orthogonalize(am_set, bra_list, ket_list, crystal):
    """
    orthogonalize atomic multipoles.

    Args:
        am_set (dict): dict of atomic multipoles in each subspace (bra_list, ket_list), { TagAtomicMultipole: NSArray }.
        bra_list ([str]): orbital list of bra.
        ket_list ([str]): orbital list of ket.
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        dict: orthogonalized multipoles.
    """

    def _ortho(**kwargs):
        t_type = kwargs.pop("t_type", None)
        tags = TagList(am_set.keys()).select(**kwargs)
        if t_type is None:
            am_set_ = {tag: m for tag, m in am_set.items() if tag in tags}
        else:
            am_set_ = {tag: m for tag, m in am_set.items() if tag in tags and tag.t_type == t_type}

        tags, mats = list(am_set_.keys()), list(am_set_.values())
        mats = NSArray(mats, style="matrix", fmt="sympy", real=False)

        if len(tags) > 0:
            mats, idx = NSArray.orthogonalize(mats)
            tags, mats = [tags[j] for j in idx], [mats[j] for j in idx]

        am_orthogonalized = {tag: NSArray(mat, style="matrix", fmt="sympy", real=False) for tag, mat in zip(tags, mats)}

        return am_orthogonalized

    b_type = basis_type(bra_list, crystal)
    if b_type == "jm":
        tags = [(tag.t_type, tag.irrep) for tag in am_set.keys()]
        tags = sorted(set(tags), key=tags.index)
        kwargs_lst = [{"t_type": t_type, "irrep": irrep} for t_type, irrep in tags]
    else:
        tags = [(tag.t_type, tag.s, tag.irrep) for tag in am_set.keys()]
        tags = sorted(set(tags), key=tags.index)
        kwargs_lst = [{"t_type": t_type, "s": s, "irrep": irrep} for t_type, s, irrep in tags]

    sub = [_ortho(**kwargs) for kwargs in kwargs_lst]
    am_set = {}
    for d in sub:
        am_set |= d

    return am_set


# ==================================================
def create_atomic_samb(bra_list, ket_list, spinful, crystal, bam, hs=None, ortho=True):
    """
    create atomic multipole basis set.

    Args:
        bra_list ([str]): orbital list of bra.
        ket_list ([str]): orbital list of ket.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.
        bam (BaseAtomicMultipoleDataset): atomic-multipole dataset for all four basis.
        hs (HarmonicsPG, optional): a set of point-group harmonics.
        parallel (bool, optional): use parallel code ?
        verbose (bool, optional): verbose parallel info ?
        ortho (bool, optional): orthogonalize ?

    Returns:
        dict: atomic SAMB, {TagMultipole: NSArray(matrix)}.
    """
    # if not spinful: 32✖️32 => 16✖️16
    U_rows = U_cols = [i for i in range(32) if i % 2 == 0]
    if not spinful:
        am_data = {tag: m[U_rows, :][:, U_cols] for tag, m in bam.items() if tag.s == 0}
    else:
        am_data = bam

    # extract subspace
    am_lm_set = _extract_subspace(am_data, bra_list, ket_list, spinful, crystal)

    # unitary transform
    if hs is None:
        am_set = am_lm_set
    else:
        am_set = _unitary_transform(am_lm_set, hs)

    # orthogonalization
    if ortho:
        am_set = _orthogonalize(am_set, bra_list, ket_list, crystal)

    return am_set



"""
HarmonicsPG manages a set of point-group harmonics (real).
"""
from multipie.harmonics.harmonics import Harmonics
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_multipole import TagMultipole
from multipie.tag.tag_list import TagList
from multipie.data.data_harmonics import (
    _data_harmonics_polar,
    _data_harmonics_axial,
)
from multipie.data.data_harmonics_real import (
    _data_harmonics_polar_real,
    _data_harmonics_axial_real,
)


# ==================================================
class HarmonicsPG(dict):  # dict of (multipole tag, harmonics), {TagMultipole: Harmonics}.
    """
    a set of point-group harmonics (real).

    Attributes:
        tag (TagGroup): point-group tag.
    """

    # ==================================================
    def __init__(self, pg_tag):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str): tag of point group.
        """
        pg_tag = TagGroup(str(pg_tag))
        self.tag = pg_tag
        """point-group tag."""

        polar = _data_harmonics_polar
        for k, v in _data_harmonics_polar_real.items():
            polar[k] = v
        axial = _data_harmonics_axial
        for k, v in _data_harmonics_axial_real.items():
            axial[k] = v

        hset = (polar, axial)

        d = {}
        for i in [0, 1]:
            for tag, hstr in hset[i][str(self.tag)].items():
                m_tag = TagMultipole(tag)
                d[m_tag] = Harmonics(tag, *hstr)

        self.update(sorted(d.items()))

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def group(self, rank, axial=False):
        """
        group for given rank and i_type.

        Args:
            rank (int): rank.
            axial (bool, optional): axial ?

        Returns:
            [(str,int)]: irreps. with given rank and i_type, (irrep., multiplicity).
        """
        head = "G" if axial else "Q"
        tags = self.key_list().select(rank=rank, head=head)
        g = list(sorted({(i.irrep, i.mul) for i in tags}))
        return g

    # ==================================================
    def select(self, **kwargs) -> list[Harmonics]:
        """
        select harmonics with given keywords.

        Args:
            kwargs (dict): select conditions for harmonics, acceptable keywords are common to TagMultipole.

        Returns:
            [Harmonics]: selected harmonics.
        """
        tags = self.key_list().select(**kwargs)
        hs = [self.get(i) for i in tags]
        hs = sorted(hs, key=lambda i: i.tag.rank)

        return hs

    # ==================================================
    def __getitem__(self, tag) -> Harmonics:
        return self.get(TagMultipole(str(tag)).to_harmonics())

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
equivalent operator.
"""
import sympy
from itertools import permutations
from gcoreutils.nsarray import NSArray


# ==================================================
def _symmetrized_polynomial(f):
    """
    convert polynomial to symmetrized sum.

    Args:
        f (sympy): polynomial.

    Returns:
        [(sympy, sympy, list)]: (coefficient, symmetrized factor, [variables])
    """
    tc = f.as_coefficients_dict()
    info = []
    for t, c in tc.items():
        tm = t.as_terms()
        s = ""
        fa = 1
        fs = 0
        for b, n in zip(tm[1], tm[0][0][1][1]):
            s += str(b) * n
            fa *= sympy.factorial(n)
            fs += n
        info.append((c, fa / sympy.factorial(fs), list(set(permutations(s)))))
    return info


# ==================================================
def _angular_momentum_matrix(j):
    """
    angular-momentum operator in desceding order of Jz.

    Args:
        j (str): magnitude of angular momentum, J (0, 1/2, 1, ...).

    Returns:
        - Matrix: Jx.
        - Matrix: Jy.
        - Matrix: Jz.
    """
    j = sympy.sympify(j)
    d = 2 * j + 1

    jz = sympy.zeros(d)
    for i in range(d):
        jz[i, i] = j - i

    jp = sympy.zeros(d)
    for i in range(1, d):
        m = j - i
        jp[i - 1, i] = sympy.sqrt((j - m) * (j + m + 1))
    jm = jp.T

    jx = (jp + jm) / 2
    jy = (jp - jm) / 2 / sympy.I

    return jx, jy, jz


# ==================================================
def equivalent_operator_from_poly(poly, j):
    """
    equivalent operator in descending order in Jz.

    Args:
        poly (sympy/str): (x,y,z) polynomial.
        j (str): magnitude of angular momentum, J (0, 1/2, 1, ...).

    Returns:
        NSArray: equivalent-operator matrix.
    """
    if type(poly) == str:
        poly = sympy.sympify(poly)
    poly = sympy.expand(poly)
    jx, jy, jz = _angular_momentum_matrix(j)
    jm = {"x": jx, "y": jy, "z": jz}
    sf = _symmetrized_polynomial(poly)

    m = sympy.zeros(jx.rows)
    for c1, c2, sv in sf:
        sf0 = sympy.zeros(jx.rows)
        for p in sv:
            sf1 = sympy.eye(jx.rows)
            for i in p:
                sf1 = sf1 * jm[i]
            sf0 += sf1
        m += c1 * c2 * sf0
    return NSArray(m.tolist(), "matrix")



"""
calculate response tensors up to 4th rank.

Notes:
    - index number starts from 1.
"""
from itertools import product
import sympy as sp
from sympy.functions.special.tensor_functions import KroneckerDelta
from sympy import LeviCivita
from multipie.data.data_compatibility_relation import _data_compatibility_relation_cubic, _data_compatibility_relation_hexagonal
from multipie.data.data_cartesian_to_ch_harmoncis import (
    _data_cartesian_to_cubic_harmonics,
    _data_cartesian_to_hexagonal_harmonics,
)

_dv = {1: "x", 2: "y", 3: "z"}
_dq = {(1, 1): "1", (2, 2): "2", (3, 3): "3", (2, 3): "4", (3, 1): "5", (1, 2): "6", (3, 2): "-4", (1, 3): "-5", (2, 1): "-6"}
_dq1 = {(1, 1): "1", (2, 2): "2", (3, 3): "3", (2, 3): "4", (3, 1): "5", (1, 2): "6", (3, 2): "4", (1, 3): "5", (2, 1): "6"}
_mul = {"a": "1", "b": "2", "c": "3", "d": "4", "e": "5", "f": "6"}


# ==================================================
def delta(i1, i2):
    """
    wrapper to ``KroneckerDelta()``

    Args:
        i1 (Number, Symbol): The first index of the KroneckerDelta function.
        i2 (Number, Symbol): The second index of the KroneckerDelta function.

    Returns:
        Number, Symbol: 1(i1==i2)/0(i1!=i2)
    """
    return KroneckerDelta(i1, i2)


# ==================================================
def epsilon(i1, i2, i3):
    """
    wrapper to ``LeviCivita()``.

    Args:
        i1 (Number, Symbol): The first index of the Levi-Civita function.
        i2 (Number, Symbol): The second index of the Levi-Civita function.
        i3 (Number, Symbol): The third index of the Levi-Civita function.

    Returns:
        Number, Symbol: 1((i1-i2)*(i2-i3)*(i3-i1)/2==1)/0(otherwise)
    """
    return LeviCivita(i1, i2, i3)


# ==================================================
def d1234(i1, i2, i3, i4):
    return delta(i1, i3) * delta(i2, i4) + delta(i1, i4) * delta(i2, i3)


# ==================================================
def d12345(i1, i2, i3, i4, i5):
    return (
        epsilon(i1, i3, i5) * delta(i2, i4)
        + epsilon(i2, i3, i5) * delta(i1, i4)
        + epsilon(i1, i4, i5) * delta(i2, i3)
        + epsilon(i2, i4, i5) * delta(i1, i3)
    )


# ==================================================
def d123456(i1, i2, i3, i4, i5, i6):
    return (
        sp.S(
            delta(i1, i3) * d1234(i2, i4, i5, i6)
            + delta(i2, i3) * d1234(i1, i4, i5, i6)
            + delta(i1, i4) * d1234(i2, i3, i5, i6)
            + delta(i2, i4) * d1234(i1, i3, i5, i6)
        )
        / 2
    )


# ==================================================
def d1234567(i1, i2, i3, i4, i5, i6, i7):
    return (
        sp.S(
            epsilon(i1, i3, i7) * d1234(i2, i4, i5, i6)
            + epsilon(i2, i3, i7) * d1234(i1, i4, i5, i6)
            + epsilon(i2, i4, i7) * d1234(i1, i3, i5, i6)
            + epsilon(i1, i4, i7) * d1234(i2, i3, i5, i6)
        )
        / 2
    )


# ==================================================
def e1234(i1, i2, i3, i4):
    return delta(i1, i3) * delta(i2, i4) - delta(i1, i4) * delta(i2, i3)


# ==================================================
def e51234(i1, i2, i3, i4, i5):
    return (
        sp.S(
            epsilon(i1, i3, i4) * delta(i2, i5)
            + epsilon(i1, i2, i4) * delta(i3, i5)
            - epsilon(i1, i2, i3) * delta(i4, i5)
            - epsilon(i2, i3, i4) * delta(i1, i5)
        )
        / 2
    )


# ==================================================
def e561234(i1, i2, i3, i4, i5, i6):
    return epsilon(i1, i2, i5) * epsilon(i3, i4, i6) + epsilon(i1, i2, i6) * epsilon(i3, i4, i5)


# ==================================================
def g12534(i1, i2, i3, i4, i5):
    return epsilon(i1, i3, i4) * delta(i2, i5) + epsilon(i2, i3, i4) * delta(i1, i5)


# ==================================================
def g125634(i1, i2, i3, i4, i5, i6):
    return (
        sp.S(
            d1234(i1, i2, i4, i5) * delta(i3, i6)
            + d1234(i1, i2, i4, i6) * delta(i3, i5)
            - d1234(i1, i2, i3, i5) * delta(i4, i6)
            - d1234(i1, i2, i3, i6) * delta(i4, i5)
        )
        / 2
    )


# ==================================================
def mp_string(lst, head, sup="", sort=True):
    """
    string of multipole

    Args:
        head (str): head of the multipole Q/G.
        sup (str): superscript.
        sort (bool, optional): sort ?

    Returns:
        str: string of multipole
    """
    lst = [_dv[i] for i in lst]
    if sort:
        s = head + "_{" + "".join(sorted("".join(lst))) + "}"
    else:
        s = head + "_{" + ("".join(lst)) + "}"
    if sup != "":
        s += "^{" + sup + "}"
    return s


# ==================================================
def mp_symbol(lst, head, sup="", sort=True):
    """
    symbol of multipole

    Args:
        head (str): head of the multipole Q/G.
        sup (str): superscript.
        sort (bool, optional): sort ?

    Returns:
        Symbol: symbol of multipole
    """
    return sp.Symbol(mp_string(lst, head, sup, sort))


# ==================================================
def S0():
    """
    component of the rank 0 tensor

    Returns:
        str: "Q_{0}"
    """
    return "Q_{0}"


# ==================================================
def S1(i1):
    """
    component of the rank 1 tensor

    Args:
        i1 (Number, Symbol): The index of the tensor.

    Returns:
        str: "Q_{i}" (i=x/y/z)
    """
    return mp_string([i1], "Q")


# ==================================================
def S12(i1, i2):
    """
    component of the symmetric part of the rank 2 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    if i1 == i2:
        lst = [(1, mp_string([i1, i2], "Q")), (1, "Q_{0}")]
    else:
        lst = [(1, mp_string([i1, i2], "Q"))]

    return lst


# ==================================================
def A12(i1, i2):
    """
    component of the asymmetric part of the rank 2 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    lst = []
    for i3 in range(1, 4):
        e = epsilon(i1, i2, i3)
        if e != 0:
            lst.append((e, mp_string([i3], "G")))

    return lst


# ==================================================
def S123(i1, i2, i3):
    """
    component of the symmetric part of the rank 3 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = [(1, mp_string([i1, i2, i3], "Q"))]
    l2 = []
    for i4, i5 in product(range(1, 4), range(1, 4)):
        g = g12534(i1, i2, i3, i4, i5)
        if g != 0:
            l2.append((g, mp_string([i4, i5], "G", "a")))
    l3 = []
    for i4 in range(1, 4):
        d = d1234(i1, i2, i3, i4)
        if d != 0:
            l3.append((d, mp_string([i4], "Q", "a")))

    if i1 == i2:
        l4 = [(1, mp_string([i3], "Q", "b"))]
    else:
        l4 = []

    return l1 + l2 + l3 + l4


# ==================================================
def A123(i1, i2, i3):
    """
    component of the asymmetric part of the rank 3 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = []
    for i4 in range(1, 4):
        e = epsilon(i1, i2, i4)
        if e != 0:
            l1.append((e, mp_string([i3, i4], "G", "b")))
    l2 = []
    for i4 in range(1, 4):
        e = e1234(i1, i2, i3, i4)
        if e != 0:
            l2.append((e, mp_string([i4], "Q", "c")))

    e = epsilon(i1, i2, i3)
    if e != 0:
        l3 = [(e, "G_{0}")]
    else:
        l3 = []

    return l1 + l2 + l3


# ==================================================
def S1234(i1, i2, i3, i4):
    """
    component of the (symmetric,symmetric,symmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = [(1, mp_string([i1, i2, i3, i4], "Q"))]
    l2 = []
    if i1 == i2:
        l2.append((1, mp_string([i3, i4], "Q", "a")))
    if i3 == i4:
        l2.append((1, mp_string([i1, i2], "Q", "a")))

    l3 = []
    for i5, i6 in product(range(1, 4), range(1, 4)):
        d = d123456(i1, i2, i3, i4, i5, i6)
        if d != 0:
            l3.append((d, mp_string([i5, i6], "Q", "b")))

    if i1 == i2 and i3 == i4:
        l4 = [(1, "Q_{0}^{a}")]
    else:
        l4 = []

    d = d1234(i1, i2, i3, i4)
    if d != 0:
        l5 = [(d, "Q_{0}^{b}")]
    else:
        l5 = []

    return l1 + l2 + l3 + l4 + l5


# ==================================================
def Sb1234(i1, i2, i3, i4):
    """
    component of the (symmetric,symmetric,asymmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = []
    for i5, i6, i7 in product(range(1, 4), range(1, 4), range(1, 4)):
        d = d1234567(i1, i2, i3, i4, i5, i6, i7)
        if d != 0:
            l1.append((d, mp_string([i5, i6, i7], "G", "a")))
    l2 = []
    if i1 == i2:
        l2.append((1, mp_string([i3, i4], "Q", "c")))
    if i3 == i4:
        l2.append((-1, mp_string([i1, i2], "Q", "c")))
    l3 = []
    for i5 in range(1, 4):
        d = d12345(i1, i2, i3, i4, i5)
        if d != 0:
            l3.append((d, mp_string([i5], "G", "a")))

    return l1 + l2 + l3


# ==================================================
def A1234(i1, i2, i3, i4):
    """
    component of the (asymmetric,asymmetric,symmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = []
    for i5, i6 in product(range(1, 4), range(1, 4)):
        e = e561234(i1, i2, i3, i4, i5, i6)
        if e != 0:
            l1.append((e, mp_string([i5, i6], "Q", "f")))

    e = e1234(i1, i2, i3, i4)
    if e != 0:
        l2 = [(e, "Q_{0}^{c}")]
    else:
        l2 = []

    return l1 + l2


# ==================================================
def Ab1234(i1, i2, i3, i4):
    """
    component of the (asymmetric,asymmetric,asymmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    lst = []
    for i5 in range(1, 4):
        e = e51234(i1, i2, i3, i4, i5)
        if e != 0:
            lst.append((e, mp_string([i5], "G", "f")))

    return lst


# ==================================================
def M1234(i1, i2, i3, i4):
    """
    component of the (symmetric,asymmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = []
    for i5 in range(1, 4):
        e = epsilon(i3, i4, i5)
        if e != 0:
            l1.append((e, mp_string([i1, i2, i5], "G", "b")))
    l2 = []
    for i5, i6 in product(range(1, 4), range(1, 4)):
        g = g125634(i1, i2, i3, i4, i5, i6)
        if g != 0:
            l2.append((g, mp_string([i5, i6], "Q", "d")))
    l3 = []
    for i5 in range(1, 4):
        g = g12534(i1, i2, i3, i4, i5)
        if g != 0:
            l3.append((g, mp_string([i5], "G", "b")))
    l4 = []
    for i5 in range(1, 4):
        de = delta(i1, i2) * epsilon(i3, i4, i5)
        if de != 0:
            l4.append((de, mp_string([i5], "G", "c")))

    return l1 + l2 + l3 + l4


# ==================================================
def Mb1234(i1, i2, i3, i4):
    """
    component of the (asymmetric,symmetric) part of the rank 4 tensor

    Args:
        i1 (Number, Symbol): The first index of the tensor.
        i2 (Number, Symbol): The second index of the tensor.
        i3 (Number, Symbol): The third index of the tensor.
        i4 (Number, Symbol): The fourth index of the tensor.

    Returns:
        list: [(coefficient, multipole)]
    """
    l1 = []
    for i5 in range(1, 4):
        e = epsilon(i1, i2, i5)
        if e != 0:
            l1.append((e, mp_string([i3, i4, i5], "G", "c")))
    l2 = []
    for i5, i6 in product(range(1, 4), range(1, 4)):
        g = g125634(i3, i4, i1, i2, i5, i6)
        if g != 0:
            l2.append((g, mp_string([i5, i6], "Q", "e")))
    l3 = []
    for i5 in range(1, 4):
        g = g12534(i3, i4, i1, i2, i5)
        if g != 0:
            l3.append((g, mp_string([i5], "G", "d")))
    l4 = []
    for i5 in range(1, 4):
        de = delta(i3, i4) * epsilon(i1, i2, i5)
        if de != 0:
            l4.append((de, mp_string([i5], "G", "e")))

    return l1 + l2 + l3 + l4


# ==================================================
def simplify_tensor(M):
    """
    simplify tensor symbol.

    ex) when Sxx = Syy, Syy is replaced by Sxx.

    Args:
        M ([list]):

    Returns:
        [list]: simplified tensor.
    """
    Xs = [x for lst in M for _, m in lst for x in m.atoms(sp.Symbol)]
    Xs = list(set(Xs))
    Cs = list(reversed([c for lst in M for c, m in lst if m != sp.S(0)]))
    args = Xs + Cs
    eqs = [sp.Eq(c, m) for lst in M for c, m in lst]
    tensor_map = sp.solve(eqs, args)

    Ms = [[None for _ in lst] for lst in M]
    for i, lst in enumerate(M):
        for j, (c, m) in enumerate(lst):
            if c in tensor_map:
                c_ = tensor_map[c]
                if c_ != sp.S(0):
                    if all([i in Cs for i in c_.atoms(sp.Symbol)]):
                        c = c_
            Ms[i][j] = (c, m)

    return Ms


# ==================================================
def create_tensor(pg_tag, tag):
    """
    create response tensors

    Args:
        pg_tag (TagGroup): tag of point group
        tag (TagResponseTensor): tag of response tensor.

    Returns:
        [list]: [[(tensor symbol, component)]]
    """
    rank, comp, i_type, t_type = tag.rank, tag.comp, tag.i_type, tag.t_type

    is_axial = i_type == "axial"
    is_magnetic_tensor = t_type == "magnetic"

    if pg_tag.is_hexagonal_subgroup():
        cr = _data_compatibility_relation_hexagonal
        mp_map = _data_cartesian_to_hexagonal_harmonics
    else:
        cr = _data_compatibility_relation_cubic
        mp_map = _data_cartesian_to_cubic_harmonics

    A_irrep = cr[str(pg_tag)][0]

    def is_A_irrep(idx):  # Is given irrep full symmetric (A) irrep ?
        return cr[str(pg_tag)][idx] == A_irrep

    def replace_head_axial(mp):
        if is_axial:
            mp = mp.translate(str.maketrans({"Q": "G", "G": "Q"}))
        return mp

    def replace_head_magnetic(mp):
        if is_magnetic_tensor:
            mp = mp.replace("Q", "T").replace("G", "M")
        return mp

    # create response tensors with full symmetric component
    # rank 0
    if rank == 0:
        mp = replace_head_axial(S0())
        c, mp, idx = mp_map[mp][0]
        x = sp.S(0)
        if is_A_irrep(idx):
            mp = replace_head_magnetic(mp)
            x += c * sp.Symbol(mp)
        M = [[(sp.Symbol(tag.latex()), x)]]
    # rank 1
    elif rank == 1:
        M = [[("", sp.S(0))] * 3] * 1
        for i1 in range(1, 4):
            x = sp.S(0)
            mp = replace_head_axial(S1(i1))
            for c, mp, idx in mp_map[mp]:
                if is_A_irrep(idx):
                    mp = replace_head_magnetic(mp)
                    x += c * sp.Symbol(mp)
            sub = _dv[i1]
            c = sp.Symbol(tag.latex() + "_{" + sub + "}")
            M[0][i1 - 1] = (c, x)
    # rank 2
    elif rank == 2:
        M = [[("", sp.S(0))] * 3 for _ in range(3)]
        if comp == "s":
            tf = S12
        else:
            tf = A12
        for i1, i2 in product(range(1, 4), range(1, 4)):
            x = sp.S(0)
            for c1, m in tf(i1, i2):
                m = replace_head_axial(m)
                for c2, mp, idx in mp_map[m]:
                    if is_A_irrep(idx):
                        mp = replace_head_magnetic(mp)
                        x += c1 * c2 * sp.Symbol(mp)
            sub = _dv[i1] + _dv[i2]
            c = sp.Symbol(tag.latex() + "_{" + sub + "}")
            M[i1 - 1][i2 - 1] = (c, x)
    # rank 3
    elif rank == 3:
        if comp == "s":
            tf = S123
            lst = [(i1, i2, i3) for i1, i2, i3 in product(range(1, 4), range(1, 4), range(1, 4)) if i1 <= i2]
            M = [[("", sp.S(0))] * 3 for _ in range(6)]
        else:
            tf = A123
            lst = [(i1, i2, i3) for i1, i2, i3 in product(range(1, 4), range(1, 4), range(1, 4)) if i1 != i2 and i1 <= i2]
            M = [[("", sp.S(0))] * 3 for _ in range(3)]
        for i1, i2, i3 in lst:
            x = sp.S(0)
            mul = ""
            for c1, m in tf(i1, i2, i3):
                if "^" in m:
                    m, mul = m.split("^")
                    mul = mul[1:-1]
                else:
                    mul = ""
                m = replace_head_axial(m)
                for c2, mp, idx in mp_map[m]:
                    if is_A_irrep(idx):
                        mp = replace_head_magnetic(mp)
                        if mul == "":
                            x += c1 * c2 * sp.Symbol(mp)
                        else:
                            x += c1 * c2 * sp.Symbol(mp + "[" + _mul[mul] + "]")

            if comp == "s":
                sub = _dq1[(i1, i2)] + _dv[i3]
                i = int(_dq1[(i1, i2)]) - 1
            else:
                sub = _dq1[(i1, i2)] + _dv[i3]
                i = int(_dq1[(i1, i2)]) - 4
                if int(_dq[(i1, i2)]) < 0:
                    x = -x
            j = i3 - 1
            c = sp.Symbol(tag.latex() + "_{" + sub + "}")
            M[i][j] = (c, x)
    # rank 4
    elif rank == 4:
        if comp == "sss":
            tf = S1234
        elif comp == "ssa":
            tf = Sb1234
        elif comp == "aas":
            tf = A1234
        elif comp == "aaa":
            tf = Ab1234
        elif comp == "sa":
            tf = M1234
        elif comp == "as":
            tf = Mb1234

        if comp in ("sss", "ssa"):
            lst = [
                (i1, i2, i3, i4)
                for i1, i2, i3, i4 in product(range(1, 4), range(1, 4), range(1, 4), range(1, 4))
                if i1 <= i2 and i3 <= i4
            ]
            M = [[("", sp.S(0))] * 6 for _ in range(6)]
        elif comp == "sa":
            lst = [
                (i1, i2, i3, i4)
                for i1, i2, i3, i4 in product(range(1, 4), range(1, 4), range(1, 4), range(1, 4))
                if i1 <= i2 and i3 <= i4 and i3 != i4
            ]
            M = [[("", sp.S(0))] * 3 for _ in range(6)]
        elif comp == "as":
            lst = [
                (i1, i2, i3, i4)
                for i1, i2, i3, i4 in product(range(1, 4), range(1, 4), range(1, 4), range(1, 4))
                if i1 <= i2 and i3 <= i4 and i1 != i2
            ]
            M = [[("", sp.S(0))] * 6 for _ in range(3)]
        else:
            lst = [
                (i1, i2, i3, i4)
                for i1, i2, i3, i4 in product(range(1, 4), range(1, 4), range(1, 4), range(1, 4))
                if i1 != i2 and i1 <= i2 and i3 != i4 and i3 <= i4
            ]
            M = [[("", sp.S(0))] * 3 for _ in range(3)]

        for i1, i2, i3, i4 in lst:
            x = sp.S(0)
            mul = ""
            for c1, m in tf(i1, i2, i3, i4):
                if "^" in m:
                    m, mul = m.split("^")
                    mul = mul[1:-1]
                else:
                    mul = ""
                m = replace_head_axial(m)
                for c2, mp, idx in mp_map[m]:
                    if is_A_irrep(idx):
                        mp = replace_head_magnetic(mp)
                        if mul == "":
                            x += c1 * c2 * sp.Symbol(mp)
                        else:
                            x += c1 * c2 * sp.Symbol(mp + "[" + _mul[mul] + "]")

            if comp in ("sss", "ssa"):
                i = int(_dq1[(i1, i2)]) - 1
                j = int(_dq1[(i3, i4)]) - 1
                sub = _dq1[(i1, i2)] + _dq1[(i3, i4)]
            elif comp == "sa":
                i = int(_dq1[(i1, i2)]) - 1
                j = int(_dq1[(i3, i4)]) - 4
                if int(_dq[(i3, i4)]) < 0:
                    x = -x
                sub = _dq1[(i1, i2)] + _dv[int(_dq1[(i3, i4)]) - 3]
            elif comp == "as":
                i = int(_dq1[(i1, i2)]) - 4
                j = int(_dq1[(i3, i4)]) - 1
                if int(_dq[(i1, i2)]) < 0:
                    x = -x
                sub = _dv[int(_dq1[(i1, i2)]) - 3] + _dq1[(i3, i4)]
            else:
                i = int(_dq1[(i1, i2)]) - 4
                j = int(_dq1[(i3, i4)]) - 4
                if int(_dq[(i1, i2)]) * int(_dq[(i3, i4)]) < 0:
                    x = -x
                sub = _dv[int(_dq1[(i1, i2)]) - 3] + _dv[int(_dq1[(i3, i4)]) - 3]

            c = sp.Symbol(tag.latex() + "_{" + sub + "}")
            M[i][j] = (c, x)

    M = [[(c, sp.simplify(m)) for c, m in lst] for lst in M]
    return M



"""
WyckoffG manages point/space-group Wyckoff positions (conventional, reduced coordinate).
"""
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_wyckoff import TagWyckoff
from multipie.tag.tag_list import TagList
from multipie.data.data_wyckoff_pg import _data_wyckoff_pg
from multipie.data.data_wyckoff_sg import _data_wyckoff_sg


# ==================================================
class WyckoffG(dict):  # dict of (wyckoff tag, wyckoff), {TagWyckoff: NSArray}.
    """
    point/space-group Wyckoff positions.

    Attributes:
        tag (TagGroup): group tag.
        v (sp.Matrix): vector variable
    """

    #
    #     __default (dict): default position for point group, {TagWyckoff: NSArray}.
    #
    # ==================================================
    def __init__(self, g_tag):
        """
        initialize the class.

        Args:
            g_tag (TagGroup or str): tag of group.
        """
        g_tag = TagGroup(str(g_tag))
        self.tag = g_tag
        """group tag."""

        self.v = NSArray.vector3d()
        """(x,y,z) vector."""

        if self.tag.is_point_group():
            self.__default = {}
            for w, pos, d in _data_wyckoff_pg[str(g_tag)]:
                v = dict(zip(["x", "y", "z"], self.v))
                self[TagWyckoff(w)] = NSArray(pos).subs(v)
                v = dict(zip(["x", "y", "z"], NSArray(d)))
                self.__default[TagWyckoff(w)] = NSArray(pos).subs(v)
        else:
            self.__default = None
            for w, pos in _data_wyckoff_sg[str(g_tag)]:
                v = dict(zip(["x", "y", "z"], self.v))
                self[TagWyckoff(w)] = NSArray(pos).subs(v)

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def position(self, wp, v=None, default=False):
        """
        wyckoff position (reduced coordinate).

        Args:
            wp (TagWyckoff or str): Wyckoff position tag.
            v (str or NSArray, optional): vector variable.
            default (bool, optional): use default value ? (point group only).

        Returns:
            NSArray: wyckoff positions.
        """
        if not self.tag.is_point_group() and default is True:
            raise NotImplementedError("default value is unsupported for space group.")

        if type(wp) == str:
            wp = TagWyckoff(wp)

        if default:
            return self.__default[wp]

        if v is None:
            return self[wp]
        else:
            v = dict(zip(["x", "y", "z"], NSArray(v)))
            return self[wp].subs(v)

    # ==================================================
    def __getitem__(self, wp):
        if type(wp) == str:
            return self.get(TagWyckoff(wp))
        else:
            return self.get(wp)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
HarmonicsPGSet manages a set of point-group harmonics (real).
"""
from multipie.harmonics.harmonics_pg import HarmonicsPG
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_list import TagList


# ==================================================
class HarmonicsPGSet(dict):  # dict of (group tag, hamonics set), {TagGroup: HarmonicsPG}.
    """
    a set of point-group harmonics (real).

    Attributes:
        tag (str): class name tag.
    """

    # ==================================================
    def __init__(self):
        """
        initialize the class.
        """
        self.tag = __class__.__name__
        """class name tag."""
        for tag in TagGroup.create():
            self[tag] = HarmonicsPG(tag)

    # ==================================================
    def __str__(self):
        return self.tag

    # ==================================================
    def __repr__(self):
        return self.tag

    # ==================================================
    def latex(self):
        return self.tag

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagGroup(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())



"""
VirtualClusterPG manages point-group virtual cluster (real).
"""
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_multipole import TagMultipole
from multipie.tag.tag_list import TagList
from multipie.data.data_virtual_cluster_real import (
    _data_virtual_cluster_basis_real,
    _data_virtual_cluster_site_real,
)


# ==================================================
class VirtualClusterPG(dict):  # dict of (multipole tag, vc basis), {TagMultipole: NSArray} (orthonormalized version).
    """
    point/space-group virtual cluster.

    Attributes:
        tag (TagGroup): group tag.
        site (NSArray): site positions (reduced coordinate).
    """

    # ==================================================
    def __init__(self, pg_tag):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str): point-group tag.
        """
        pg_tag = TagGroup(pg_tag)
        self.tag = pg_tag
        """point-group tag."""

        assert self.tag.is_point_group(), "it is valid only for point group."

        vc = _data_virtual_cluster_basis_real
        vc_site = _data_virtual_cluster_site_real

        self._basis = {}
        self._site = {}
        for wp, dic in vc[str(pg_tag)].items():
            self._basis[wp] = {}
            for m, b in dic.items():
                self._basis[wp][TagMultipole(m)] = NSArray(b)
        for wp, vec in vc_site[str(pg_tag)].items():
            self._site[wp] = NSArray(vec)

        wp0 = list(self._basis.keys())[0]  # assume general point is the first.
        d = {}
        for m, b in self._basis[wp0].items():
            d[m] = b
        self.update(d)
        self.site = self._site[wp0]
        """site positions (cartesian coordinate)."""

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagMultipole(tag))
        else:
            return self.get(tag)

    # ==================================================
    def get(self, wyckoff, tag):
        if type(tag) == str:
            return self._basis[wyckoff][TagMultipole(tag)]
        else:
            return self._basis[wyckoff][tag]

    # ==================================================
    def key_list(self, wyckoff=None):
        if wyckoff is None:
            return TagList(self.keys())
        else:
            return TagList(self._basis[wyckoff].keys())



"""
construct model from matrx dict.
"""
import numpy as np
from gcoreutils.nsarray import NSArray


# ==================================================
def construct_samb_matrix(matrix_dict, k_grid=None):
    """
    construct SAMB matrix from matrix dict.

    Args:
        matrix_dict (dict): matrix dict from _matrix.py file.
        k_grid (NSArray, optional): k_grid points (reduced coordinate).

    Returns:
        ndarray: basis matrix, [#bases, #k, dim, dim].
    """
    z_samb = matrix_dict["matrix"]
    z_mat = {k: str(NSArray(v).evalf()) for k, v in z_samb.items()}

    if matrix_dict["molecule"] or k_grid is None:
        Nk = 1
        bond = [{}]
    else:
        Nk = len(k_grid)
        d = {}
        for name, ex in matrix_dict["bond"].items():
            ex = NSArray(ex, fmt="value")
            ex = np.broadcast_to(ex, (Nk, 3))
            k_grid = np.array(k_grid)
            p = 2 * np.pi * np.sum(ex * np.array(k_grid), axis=1)
            d[name.replace("bond_", "c")] = np.cos(p)
            d[name.replace("bond_", "s")] = np.sin(p)

        bond = [{name: lst[i] for name, lst in d.items()} for i in range(len(k_grid))]

    def multiple_replace(s, d):
        s_ = s
        [s_ := s_.replace(name, f"({val})") for name, val in d.items()]
        return s_

    vmultiple_replace = np.vectorize(multiple_replace)

    Z_list = [Zj.replace("I", "1j") for Zj in z_mat.values()]
    Z = np.array([np.array([np.array(eval(m), dtype="complex64") for m in vmultiple_replace(Zj, bond)]) for Zj in Z_list])

    return Z


# ==================================================
def convert_samb_to_matrix_set(Z, param=None):
    """
    convert SAMB to a set of matrix.

    Args:
        Z (ndarray): samb matrix, [#bases, #k, dim, dim].
        param (list, optional): parameter set(s), [z_j] for crystal or [z_j]/[[z_j]] for molecule.

    Returns:
        ndarray: matrix, [#params/#k, dim, dim].
    """
    if param is None:
        param = [1.0] * Z.shape[0]
    param = np.array(param)
    if param.ndim == 1:
        M = [Z.transpose(1, 2, 3, 0) @ param]
    else:
        M = [Z.transpose(1, 2, 3, 0) @ p for p in param]
    M = np.vstack(M)

    return M



"""
This file provides utility functions for atomic orbital.
"""
import sympy as sp
from collections import Counter

from gcoreutils.string_util import remove_space
from multipie.multipole.util.spin_orbital_basis import (
    _standard_basis,
    _abbrev_basis,
)
from multipie.const import __def_dict__


# ==================================================
def _remove_space(orb_list):
    """
    remove space

    Args:
        orb_list (str/[str]): orbital.

    Returns:
        list: orbital list ?
    """
    if type(orb_list) == str:
        return [o for o in orb_list.split(" ") if o != ""]
    elif type(orb_list) == list:
        return [remove_space(str(o)) for o in orb_list]
    else:
        raise Exception(f"invalid type of orbital list = {type(orb_list)} is given.")


# ==================================================
def is_spinful(orb_list):
    """
    spinful ?

    Args:
        orb_list (str/[str]): orbital.

    Returns:
        bool: spinful ?
    """
    orb_list = _remove_space(orb_list)
    if all([len(o) > 2 and o[2] == "/" for o in orb_list]):  # jm basis
        return True
    elif all(["U" in o or "D" in o for o in orb_list]):  # spinful basis
        return True
    else:
        return False


# ==================================================
def to_spinful(orb_list):
    """
    convert from spinless orbital list to spinful orbital list.

    Args:
        orb_list (str/[str]): orbital list.

    Returns:
        list: spinful orbital list.
    """
    orb_list = _remove_space(orb_list)

    if all([is_spinful(o) for o in orb_list]):  # spinful basis
        return orb_list
    else:
        sf_orb_list = []
        if all([any([k in o for k in ("s", "p", "d", "f")]) for o in orb_list]):
            for o in orb_list:
                sf_orb_list.append(f"({o},U)")
                sf_orb_list.append(f"({o},D)")
        else:  # lm
            for o in orb_list:
                sf_orb_list.append(f"({o[1:-1]},U)")
                sf_orb_list.append(f"({o[1:-1]},D)")

        return sf_orb_list


# ==================================================
def to_spinless(orb_list):
    """
    convert from spinful orbital list to spinless orbital list.

    Args:
        orb_list (str/[str]): orbital list.

    Returns:
        list: spinless orbital list.
    """
    orb_list = _remove_space(orb_list)

    if all([len(o) > 2 and o[2] == "/" for o in orb_list]):  # jm
        return orb_list
    elif not all(["U" in o or "D" in o for o in orb_list]):  # spinless basis
        return orb_list
    else:
        sl_orb_list = []
        for o in orb_list:
            for b_type, sb in _standard_basis[1].items():
                if o not in sb:
                    continue
                idx = sb.index(o)
                idx = sp.S(idx) / 2 if idx % 2 == 0 else sp.S(idx - 1) / 2
                sl_orb_list.append(_standard_basis[0][b_type][idx])

        sl_orb_list = list(sorted(sl_orb_list, key=sl_orb_list.index))
        return sl_orb_list


# ==================================================
def basis_type(orb_list, crystal):
    """
    identify the type of basis.

    Args:
        orb_list (str/[str]): orbital list.
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        (tuple/NoneType): type of basis (jm/lm/cubic/hexagonal).
    """
    orb_list = _remove_space(orb_list)
    orb_list = to_spinless(orb_list)

    if all(["/" in o for o in orb_list]):
        return "jm"
    elif all([any([k in o for k in ("s", "p", "d", "f")]) for o in orb_list]):  # cubic/hexagonal
        if crystal in ("triclinic", "monoclinic", "orthorhombic", "tetragonal", "cubic"):
            return "cubic"
        elif crystal in ("trigonal", "hexagonal"):
            return "hexagonal"
        else:
            raise Exception("invalid crystal = {crystal} is given.")
    else:
        return "lm"


# ==================================================
def to_standard_form(orb_list, crystal):
    """
    convert from an abbreviation form to a standard form.

    Args:
        orb_list (str/[str]): orbital list.
        b_type (str): type of basis, "lm/jm/cubic/hexagonal".

    Returns:
        list: orbital list in standard format.

    Note:
        ex) ["p"] -> ["px", "py", "pz"].
    """
    orb_list = _remove_space(orb_list)

    b_type = basis_type(orb_list, crystal)
    spinful = 1 if b_type == "jm" else 0

    standard_orb_list = []
    for o in orb_list:
        if o in _abbrev_basis[spinful][b_type]:
            if b_type in ("cubic", "hexagonal"):
                orbs = list(filter(lambda b: o in b, _standard_basis[0][b_type]))
            elif b_type == "lm":
                orbs = list(filter(lambda b: o == b[1], _standard_basis[0]["lm"]))
            elif b_type == "jm":
                orbs = list(filter(lambda b: o == "(" + b[1:4] + "," + b[-2] + ")", _standard_basis[1]["jm"]))
            else:
                raise Exception(f"invalid b_type = {b_type} is given.")
            standard_orb_list.extend(orbs)
        else:
            standard_orb_list.append(o)

    return standard_orb_list


# ==================================================
def invalid_orb_list(orb_list):
    """
    list of orbital list with the invalid forms included in given orbital list.

    Args:
        orb_list (str/[str]): orbital list.

    Returns:
        [str]: invalid orbital list.
    """
    orb_list = _remove_space(orb_list)

    orb_list_dataset = (
        sum(list(_standard_basis[0].values()), [])
        + sum(list(_standard_basis[1].values()), [])
        + sum(list(_abbrev_basis[0].values()), [])
        + sum(list(_abbrev_basis[1].values()), [])
    )

    orb_list_ = set(orb_list) & set(orb_list_dataset)
    invalid_orbs = set(orb_list) - orb_list_

    return invalid_orbs


# ==================================================
def invalid_f_orb_list(orb_list, crystal):
    """
    list of invalid f orbital list included in given orbital list.

    Args:
        orb_list (str/[str]): orbital list.
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        [str]: invalid f orbital list.
    """
    orb_list = _remove_space(orb_list)

    # f orb_list specific to cubic and hexagonal basis.
    f_orb_dict = {
        "cubic": ("fax", "fay", "fbx", "fby"),
        "hexagonal": ("f1", "f2", "fx", "fy"),
    }
    f_orb_list = f_orb_dict["cubic"] + f_orb_dict["hexagonal"]

    # correspondence between the seven crystal systems and harmonics types.
    invalid_f_orbs = []
    for o in orb_list:
        if o in f_orb_list:
            b_type = "cubic" if o in f_orb_dict["cubic"] else "hexagonal"
            is_f_orb = o in f_orb_dict[b_type]
            is_parent_crystal = crystal in __def_dict__["parentgroup"][b_type]
            if is_f_orb and is_parent_crystal:
                continue

            invalid_f_orbs.append(o)

    return invalid_f_orbs


# ==================================================
def is_insufficient(orb_list, crystal):
    """
    check if same irrep pair is not included in given orbital list.

    Args:
        orb_list (str/[str]): orbital list.
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        bool, str: same irrep pair is not included in given orbital list ?, orbit list that must be included at the same time.
    """
    orb_list = _remove_space(orb_list)

    pdf_orb_list = {
        "p": {o for o in orb_list if o[0] == "p"},
        "d": {o for o in orb_list if o[0] == "d"},
        "f": {o for o in orb_list if o[0] == "f"},
    }

    pdf_orb_dict = {
        "cubic": [
            ("px", "py", "pz"),
            ("du", "dv"),
            ("dyz", "dzx", "dxy"),
            ("fax", "fay", "faz"),
            ("fbx", "fby", "fbz"),
        ],
        "tetragonal": [("px", "py"), ("dyz", "dzx"), ("fax", "fay"), ("fbx", "fby")],
        "hexagonal": [("px", "py"), ("dxy", "dv"), ("fx", "fy"), ("fbz", "fxyz"), ("fax", "fay"), ("fbx", "fby")],
        "trigonal": [("px", "py"), ("dxy", "dv"), ("fx", "fy"), ("fbz", "fxyz"), ("fax", "fay"), ("fbx", "fby")],
        "triclinic": [],
        "monoclinic": [],
        "orthorhombic": [],
    }

    orbs_list = pdf_orb_dict[crystal]
    for orbs in orbs_list:
        o = pdf_orb_list[orbs[0][0]]
        if not o.isdisjoint(set(orbs)) and not o.issuperset(set(orbs)):
            return True, tuple(orbs)

    return False, ""


# ==================================================
def rank(o, spinful, crystal):
    """
    rank of given orbital.

    Args:
        o (str): orbital.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        int: rank.
    """
    b_type = basis_type(o, crystal)
    if b_type == "jm":
        return sp.sympify(o)[2]
    elif b_type == "lm":
        return sp.sympify(o)[0]
    elif b_type in ("cubic", "hexagonal"):
        spdf_rank = {"s": 0, "p": 1, "d": 2, "f": 3}
        return spdf_rank[o[1]] if spinful else spdf_rank[o[0]]
    else:
        raise Exception(f"invalid type of basis = {b_type} is given.")


# ==================================================
def parse_orb_list(orb_list, spinful, crystal):
    """
    parse given orbital list.
    convert orbital list from an abbreviation form to a standard form.

    Args:
        orb_list (str/[str]): orbital list.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        ([str]): orbital list.
    """
    if type(orb_list) == str:  # abbreviation form
        orb_list = [o for o in orb_list.split(" ") if o != ""]
        is_abbrev = True
    elif type(orb_list) == list:  # standard form
        orb_list = [remove_space(str(o)) for o in orb_list]
        is_abbrev = False
    else:
        raise Exception(f"invalid type of orb_list = {type(orb_list)} is given.")

    # check if invalid form of orbital is included in the given orb_list.
    d = Counter(orb_list)
    is_duplicated = not all(list(d.values()))
    if is_duplicated:
        raise Exception(f"duplication is included in given orb_list, {d}")

    # check if invalid orb_list are included.
    invalid_orbs = invalid_orb_list(orb_list)
    is_invalid_orbs = len(invalid_orbs) != 0
    if is_invalid_orbs:
        raise Exception(f"invalid orb_list = {invalid_orbs} is given.")

    # if one of ("fax", "fay", "fbx", "fby", "f1", "f2", "fx", "fy") is given explicitly, check consistency with crystal.
    invalid_f_orbs = invalid_f_orb_list(orb_list, crystal)
    is_invalid_f_orbs = len(invalid_f_orbs) != 0
    if is_invalid_f_orbs:
        raise Exception(f"invalid f orb_list = {invalid_f_orbs} is given.")

    # check if same irrep pair is included in given orb_list.
    if crystal in ("triclinic", "monoclinic", "orthorhombic", "tetragonal", "trigonal", "hexagonal", "cubic"):
        flag, orbs = is_insufficient(orb_list, crystal)
        if flag:
            raise Exception(f"orb_list must include same irrep pairs, {orbs}, but {orb_list} was fiven.")

    # convert from an abbreviation form to a standard form.
    if is_abbrev:
        orb_list = to_standard_form(orb_list, crystal)

    # convert from spinless to spinful basis set.
    if spinful:
        orb_list = to_spinful(orb_list)

    return orb_list


# ==================================================
def sort_orb_list(orb_list, spinful, crystal):
    """
    sort given orbital list.

    Args:
        orb_list (str/[str]): orbital list.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        ([str]): sorted orbital list.
                ["px", "py", "s"] => ["s", "px", "py"]
    """
    if type(orb_list) == str:  # abbreviation form
        orb_list = [o for o in orb_list.split(" ") if o != ""]
    elif type(orb_list) == list:  # standard form
        orb_list = [remove_space(str(o)) for o in orb_list]
    else:
        raise Exception(f"invalid type of orb_list = {type(orb_list)} is given.")

    orb_list = to_spinless(orb_list)

    b_type = basis_type(orb_list, crystal)
    spinful_ = True if b_type == "jm" else False

    orb_list_data = _standard_basis[spinful_][b_type]

    max_l = 3
    orb_list_sorted = [o for l in range(max_l + 1) for o in orb_list if l == rank(o, spinful_, crystal)]
    orb_list_sorted = [o for o in orb_list_data if o in orb_list_sorted]

    if spinful:
        orb_list_sorted = to_spinful(orb_list_sorted)

    return orb_list_sorted


# ==================================================
def split_orb_list_rank_block(orb_list, spinful, crystal):
    """
    split given orbital list into rank block.

    Args:
        orb_list (str/[str]): orbital list.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        [[str]]: splitted orbital list.
                ["s","px","py"] => [["s"], ["px","py"]].
    """
    max_l = 3
    splitted_orb_list = [[o for o in orb_list if l == rank(o, spinful, crystal)] for l in range(max_l + 1)]
    splitted_orb_list = [lst for lst in splitted_orb_list if len(lst) > 0]

    return splitted_orb_list


# ==================================================
def to_latex(orb_list, spinful, crystal):
    """
    convert string to latex format.

    Args:
        orb_list (str/[str]): orbital list.
        spinful (bool): spinful ?
        crystal (str): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.

    Returns:
        list: orbital in latex format.
    """

    def f(orb):
        if spinful:
            orb, UD = orb[1:-1].split(",")
            if len(orb) > 1:
                head, sub = orb[0], orb[1:]
                orb = head + "_{" + sub + "}"
            orb = f"({orb},{UD})"
        else:
            if len(orb) > 1:
                head, sub = orb[0], orb[1:]
                orb = head + "_{" + sub + "}"
        return orb

    b_type = basis_type(orb_list, crystal)

    if b_type in "jm":
        return orb_list
    elif b_type in "lm":
        return orb_list
    else:
        if type(orb_list) == str:
            orb_list = f(orb_list)
        elif type(orb_list) == list:
            orb_list = [f(orb) for orb in orb_list]
        return orb_list



"""
PointGroup manages point group.
"""
import sympy as sp
import numpy as np
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_wyckoff import TagWyckoff
from multipie.tag.tag_list import TagList
from multipie.tag.tag_multipole import TagMultipole
from multipie.symmetry_operation.util.symmetry_operation_util import to_cartesian
from multipie.symmetry_operation.symmetry_operation_g_set import SymmetryOperationGSet
from multipie.symmetry_operation.symmetry_operation_g import SymmetryOperationG
from multipie.wyckoff.wyckoff_g_set import WyckoffGSet
from multipie.wyckoff.wyckoff_g import WyckoffG
from multipie.harmonics.harmonics_pg_set import HarmonicsPGSet
from multipie.harmonics.harmonics_pg import HarmonicsPG
from multipie.character.character_pg_set import CharacterPGSet
from multipie.character.character_pg import CharacterPG
from multipie.virtual_cluster.virtual_cluster_pg_set import VirtualClusterPGSet
from multipie.virtual_cluster.virtual_cluster_pg import VirtualClusterPG
from multipie.clebsch_gordan.clebsch_gordan_pg_set import ClebschGordanPGSet
from multipie.clebsch_gordan.clebsch_gordan_pg import ClebschGordanPG
from multipie.response_tensor.response_tensor_pg_set import ResponseTensorPGSet
from multipie.response_tensor.response_tensor_pg import ResponseTensorPG
from multipie.multipole.base.base_atomic_multipole_dataset import BaseAtomicMultipoleDataset
from multipie.multipole.util.atomic_samb_util import create_atomic_samb
from multipie.multipole.util.atomic_orbital_util import basis_type
from multipie.multipole.util.z_samb_util import create_z_samb
from multipie import get_binary


# ==================================================
class PointGroup:
    """
    point group.

    Attributes:
        tag (TagGroup): point-group tag.
        symmetry_operation (SymmetryOperationG): the class to manage symmetry operations.
        wyckoff (WyckoffG): the class to manage Wyckoff positions.
        harmonics (HarmonicsPG): the class to manage harmonics.
        character (CharacterPG): the class to manage character table.
        virtual_cluster (VirtualClusterPG): the class to manage virtual cluster.
        clebsch_gordan (ClebschGordanPG): the class to manage Clebsch-Gordan coefficients.
        response (ResponseTensorPG) : the class to manage response tensors.
        mp_dataset (BaseAtomicMultipoleDataset) : the dict to manage atomic multipole matrices.
        core (BinaryManager): core for binary data.
    """

    # ==================================================
    def __init__(self, pg_tag=None, core=None, verbose=False):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str, optional): point-group tag.
            core (BinaryManager, optional): core for binary data.
            verbose (bool, optional): verbose access to binary data ?

        Notes:
            - if pg_tag is None, "C1" is used.
        """
        if pg_tag is None:
            pg_tag = "C1"
        if type(pg_tag) == str:
            pg_tag = TagGroup(pg_tag)
        assert pg_tag in TagGroup.create(), f"{pg_tag} is not a point-group tag."

        if core is None:
            core = get_binary(verbose=verbose)

        self.tag = pg_tag
        """point-group tag."""

        self.harmonics: HarmonicsPG = core[HarmonicsPGSet][pg_tag]
        """the class to manage harmonics."""

        self.character: CharacterPG = core[CharacterPGSet][pg_tag]
        """the class to manage character table."""

        self.symmetry_operation: SymmetryOperationG = core[SymmetryOperationGSet][pg_tag]
        """the class to manage symmetry operations."""

        self.wyckoff: WyckoffG = core[WyckoffGSet][pg_tag]
        """the class to manage Wyckoff positions."""

        self.virtual_cluster: VirtualClusterPG = core[VirtualClusterPGSet][pg_tag]
        """the class to manage virtual cluster."""

        self.clebsch_gordan: ClebschGordanPG = core[ClebschGordanPGSet][pg_tag]
        """the class to manage Clebsch-Gordan coefficients."""

        self.response: ResponseTensorPG = core[ResponseTensorPGSet][pg_tag]
        """the class to manage response tensors."""

        self.mp_dataset: BaseAtomicMultipoleDataset = core[BaseAtomicMultipoleDataset]
        """the class to manage atomic multipole matrices."""

        self.core = core
        """binary data."""

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    @property
    def identity_irrep(self):
        """
        identity irreducible representation.

        Returns:
            str: identity irreducible representation.
        """
        return str(self.character.irrep_list[0])

    # ==================================================
    def irrep_decomposition_table(self):
        """
        product irreducible decomposition table.

        Returns:
            dict: (irrep1,irrep2): {irrep:n}, { (str, str): [(int,str)] }.
        """
        dic = {}
        for irrep1 in self.character.key_list():
            for irrep2 in self.character.key_list():
                d = {str(ir): n for n, ir in self.character.symmetric_product_decomposition((irrep1, irrep2), False)}
                if irrep1 == irrep2:
                    for n, ir in self.character.anti_symmetric_product_decomposition(irrep1, False):
                        d[str(ir)] = d.get(str(ir), 0) + n
                dic[(str(irrep1), str(irrep2))] = d

        return dic

    # ==================================================
    def transform_matrix_site(self, site, cc_only=False):
        """
        trasform matrix for site basis (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".
            cc_only (bool, optional): conjugacy-class operations only ?

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: site basis.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        s = self.symmetry_operation._equivalent_vector(site, remove_duplicate=True)
        op = self.symmetry_operation.mat(cc_only=cc_only)

        tm = []
        for ii in range(len(op)):
            opi = op[ii]
            tmg = np.zeros((len(s), len(s))).astype(int).tolist()
            for i in range(len(s)):
                si = s[i].transform(opi)
                tmg[s.index(si)][i] = 1
            tm.append(tmg)

        tm = NSArray(tm, "matrix")

        return tm, s

    # ==================================================
    def transform_matrix_bond(self, bond, nondirectional=False, cc_only=False):
        """
        trasform matrix for bond basis (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            nondirectional (bool, optional): ignore directional property ?
            cc_only (bool, optional): conjugacy-class operations only ?

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: bond basis.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        b = self.symmetry_operation._equivalent_bond(bond, nondirectional=nondirectional, remove_duplicate=True)
        op = self.symmetry_operation.mat(cc_only=cc_only)

        tm = []
        for ii in range(len(op)):
            opi = op[ii]
            tmg = np.zeros((len(b), len(b))).astype(int).tolist()
            for i in range(len(b)):
                bi = b[i].transform(opi)
                if nondirectional:
                    j = b.index(bi)
                    if j is None:
                        j = b.index(bi.reverse_direction())
                else:
                    j = b.index(bi)
                tmg[j][i] = 1
            tm.append(tmg)

        tm = NSArray(tm, "matrix")

        return tm, b

    # ==================================================
    def transform_matrix_vector(self, axial=False, cc_only=False):
        """
        trasform matrix for vector basis (cartesian coordinate).

        Args:
            axial (bool, optional); axial vector ?
            cc_only (bool, optional): conjugacy-class operations only ?

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: vector basis.
        """
        tm = self.symmetry_operation.mat(cc_only=cc_only, axial=axial, cartesian=True)
        head = "G" if axial else "Q"
        v = NSArray.vector3d(head=head)

        return tm, v

    # ==================================================
    def transform_matrix_orbital(self, axial=False, cc_only=False, **kwargs):
        """
        transform matrix for harmonics basis (cartesian coordinate).

        Args:
            axial (bool, optional): axial harmonics ?
            cc_only (bool, optional): conjugacy-class operations only ?
            kwargs (dict, optional): select conditions for harmonics, (rank/irrep/mul/comp).

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: harmonics basis.
        """
        head = "G" if axial else "Q"
        orbitals = self.harmonics.select(head=head, **kwargs)
        v = NSArray.vector3d(head=head)
        t_pos = self.symmetry_operation._equivalent_vector(v, cc_only=cc_only, axial=False, cartesian=True)

        hset = NSArray([i.expression(v=v) for i in orbitals], "scalar")

        if axial:
            p_det = self.symmetry_operation.mat(axial=False, cc_only=cc_only).det()
            tm = self.symmetry_operation._transform_matrix_orbital_basis(v, t_pos, hset, p_det)
        else:
            tm = self.symmetry_operation._transform_matrix_orbital_basis(v, t_pos, hset)

        return tm, hset

    # ==================================================
    def transform_site(self, site, remove_duplicate=False):
        """
        set of transformed sites (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".
            remove_duplicate (bool, optional): remove the same sites ?

        Returns:
            NSArray: transformed sites (in order of symmetry operations).

        Notes:
            - when remove_duplicate is True, return values in the same order as site_mapping keys.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        if remove_duplicate:
            s = NSArray.from_str(self.site_mapping(site).keys())
        else:
            s = self.symmetry_operation._equivalent_vector(site, remove_duplicate=False)

        return s

    # ==================================================
    def transform_bond(self, bond, nondirectional=False, remove_duplicate=False):
        """
        set of transformed bonds (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            nondirectional (bool, optional): ignore directional property ?
            remove_duplicate (bool, optional): remove the same bonds ?

        Returns:
            NSArray: transformed bonds (in order of symmetry operations).

        Notes:
            - when remove_duplicate and nondirectional are True, return values in the same order as bond_mapping keys.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        if remove_duplicate and nondirectional:
            b = NSArray.from_str(self.bond_mapping(bond)[0].keys())
        else:
            b = self.symmetry_operation._equivalent_bond(bond, nondirectional=nondirectional, remove_duplicate=remove_duplicate)

        return b

    # ==================================================
    def transform_vector(self, vector, axial=False):
        """
        set of transformed vectors (cartesian coordinate).

        Args:
            vector (str or NSArray): vector, "[x,y,z]".
            axial (bool, optional); axial vector ?

        Returns:
            NSArray: transformed vectors (in order of symmetry operations).
        """
        if not isinstance(vector, (str, NSArray)):
            raise KeyError(f"{type(vector)} is not accepted for vector.")

        if type(vector) == str:
            vector = NSArray(vector)
        op = self.symmetry_operation.mat(cc_only=False, axial=axial, cartesian=True)
        vector = op.apply(vector)

        return vector

    # ==================================================
    def transform_orbital(self, orbital, axial=False):
        """
        set of transformed orbitals (cartesian coordinate).

        Args:
            orbital (str): (xyz)-polynomial expression.
            axial (bool, optional): axial orbital ?

        Returns:
            NSArray: transformed orbitals (in order of symmetry operations).
        """
        v = NSArray.vector3d()
        tm = self.symmetry_operation.mat(cc_only=False, axial=False, cartesian=True).inverse()
        tr = tm.apply(v)

        h = NSArray(orbital).subs({"x": v[0], "y": v[1], "z": v[2]})
        if axial:
            so = self.symmetry_operation.mat(axial=False, cc_only=False).det()
            t_orb = [(so[i] * self.symmetry_operation._transform_variable(h, v, tr[i])).tolist() for i in range(len(tr))]
        else:
            t_orb = [self.symmetry_operation._transform_variable(h, v, tr[i]).tolist() for i in range(len(tr))]

        return t_orb

    # ==================================================
    def irrep_decomposition(self, point_group, rank, axial=False):
        """
        irrep. decomposition of harmonics in terms of given point group.

        Args:
            point_group (str): point-group tag to be decompsed.
            rank (int): rank of harmonics.
            axial (bool, optional): axia harmonics ?

        Returns:
            list: decomposition info., [(harm_tag, [(coeff,harmonics basis of this group)].
        """
        head = "G" if axial else "Q"

        hset = self.harmonics.select(rank=rank, head=head)  # self

        pg = PointGroup(point_group, core=self.core)
        other = pg.harmonics.select(rank=rank, head=head)  # other

        decomp = []
        for h in hset:
            Uself = h.u_matrix()
            c = []
            for o in other:
                Uother = o.u_matrix()
                ci = NSArray.dot(Uother, Uself)
                if ci != 0:
                    c.append((ci, str(o)))
            decomp.append((str(h), c))

        return decomp

    # ==================================================
    def find_rep_site(self, site_list):
        """
        find representative sites from site_list.

        Args:
            site_list (NSArray): site list.

        Returns:
            NSArray: representative sites.
        """

        def remove_equivalent_site(sl, rep_site):
            if not sl:
                return rep_site
            else:
                s = list(sl)[0]
                r = set(self.transform_site(s, remove_duplicate=True).str())
                rep_site.append(s)
                return remove_equivalent_site(sl - r, rep_site)

        assert len(site_list) > 0, "empty site list is given."

        site_list = set(site_list.str())

        rep_site = []
        rep_site = remove_equivalent_site(site_list, rep_site)
        rep_site = NSArray.from_str(rep_site)

        return rep_site

    # ==================================================
    def find_rep_bond(self, bond_list):
        """
        find representative bonds from bond_list.

        Args:
            bond_list (NSArray): bond list.

        Returns:
            NSArray: representative bonds.
        """

        def remove_equivalent_bond(bl, rep_bond):
            if not bl:
                return rep_bond
            else:
                b = list(bl)[0]
                r = self.transform_bond(b, remove_duplicate=True)
                r = NSArray.concat([r, r.reverse_direction()])
                r = set(r.str())
                rep_bond.append(b)
                return remove_equivalent_bond(bl - r, rep_bond)

        assert len(bond_list) > 0, "empty bond list is given."

        # remove equivalent bonds.
        tail = bond_list[0].convert_bond("bond_th")[0]
        bond_list = set(bond_list.str())

        # find representative bonds.
        rep_bond = []
        rep_bond = remove_equivalent_bond(bond_list, rep_bond)
        rep_bond = NSArray.from_str(rep_bond)

        # set bonds with given tail.
        rep_bond1 = []
        for bond in rep_bond:
            bonds = self.transform_bond(bond, remove_duplicate=True)
            t, h = bonds.convert_bond("bond_th")
            idx = t.index(tail)
            if idx is None:
                idx = h.index(tail)
                assert idx is not None, "not reach here."
                b = bonds[idx].reverse_direction()
            else:
                b = bonds[idx]
            v, c = b.convert_bond("bond")
            rep_bond1.append(f"{v}@{c}")
        rep_bond = NSArray.from_str(rep_bond1)

        return rep_bond

    # ==================================================
    def site_mapping(self, site):
        """
        mapping between site and symmetry operations (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".

        Returns:
            dict: mapping from site string to symmetry-operation IDs.

        Notes:
            - sites are sorted by 1st SO.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        s = self.symmetry_operation._equivalent_vector(site)

        lst = {}
        for i in range(len(s)):
            si = str(s[i])
            lst[si] = lst.get(si, []) + [i]

        # sort in order of 1st component of SO.
        lst = dict(sorted(lst.items(), key=lambda i: i[1][0]))

        return lst

    # ==================================================
    def bond_mapping(self, bond):
        """
        mapping between bond and symmetry operations (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".

        Returns: tuple
            - dict: mapping from bond to symmetry-operation IDs. bond = "tail;head".
            - bool: nondirectional ?

        Notes:
            - if bond direction is reversed, symmetry operation ID with minus sign.
            - bonds are sorted by 1st SO.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")
        if type(bond) == str:
            bond = NSArray(bond)

        b_d = self.symmetry_operation._equivalent_bond(bond, nondirectional=False, remove_duplicate=True)
        b_all = self.symmetry_operation._equivalent_bond(bond, nondirectional=False, remove_duplicate=False)
        b_nd = self.symmetry_operation._equivalent_bond(bond, nondirectional=True, remove_duplicate=True)

        nd = len(b_nd) != len(b_d)
        b = b_nd if nd else b_d

        tail0 = bond.convert_bond("bond_th")[0]
        tail0 = NSArray.from_str(self.site_mapping(tail0).keys())
        head0 = bond.convert_bond("bond_th")[1]
        head0 = NSArray.from_str(self.site_mapping(head0).keys())

        # set bond direction with the same tail as given one.
        if str(tail0.sort()) == str(head0.sort()):
            for i in range(len(b)):
                t, h = b[i].convert_bond("bond_th")
                t_idx = tail0.index(t)
                h_idx = tail0.index(h)
                if t_idx > h_idx:
                    b[i] = b[i].reverse_direction()

        lst1 = {}
        for i in range(len(b_all)):
            bi = b_all[i]
            bib = str(bi.reverse_direction())
            bi = str(bi)
            if b.index(bi) is not None:
                lst1[bi] = lst1.get(bi, []) + [i + 1]
            elif b.index(bib) is not None:
                lst1[bib] = lst1.get(bib, []) + [-(i + 1)]
            else:
                assert False, "not reach here."

        # reverse SO if identity SO idx are negative.
        if sum(list(lst1.values()), []).count(-1) > 0:
            lst = {i: [-k for k in j] for i, j in lst1.items()}
        else:
            lst = lst1

        # reset SO number staring from zero.
        lsts = {}
        for i, j in lst.items():
            lsts[i] = [k - 1 if k > 0 else k + 1 for k in j]

        # sort in order of 1st component of SO.
        lsts = dict(sorted(lsts.items(), key=lambda i: abs(i[1][0])))

        return lsts, nd

    # ==================================================
    def find_wyckoff_position(self, site):
        """
        find Wyckoff position for a given site (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".

        Returns:
            str: Wyckoff position.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        wpv = self.symmetry_operation._equivalent_vector(site, remove_duplicate=True)
        n = len(wpv)
        tags = [i for i in self.wyckoff.keys() if i.n == n]
        if len(tags) == 1:
            return tags[0]

        wps = [self.wyckoff.position(tag) for tag in tags]
        w0 = wpv[0]
        for tag, wp in zip(tags, wps):
            for i in wp:
                if str(i) == str(w0):
                    return tag
                if sp.solve(list(i - w0), self.wyckoff.v.tolist(), manual=True):
                    return tag

        raise Exception(f"{str(site)} cannot be found.")

    # ==================================================
    def virtual_cluster_basis(self, wyckoff=None, ortho=True, create=False):
        """
        virtual-cluster basis set.

        Args:
            wyckoff (str, optional): Wyckoff position.
            ortho (bool, optional): orthogonalize ?
            create (bool, optional): create from scratch ?

        Returns: tuple
            - TagMultipole or NSArray: virtual-cluster basis set, (multipole tag, basis vector).
            - NSArray: site basis (reduced coordinate).
        """
        if wyckoff is not None and TagWyckoff(wyckoff) in self.wyckoff.keys():
            wp = wyckoff
        else:
            wp = self.wyckoff.key_list()[-1]

        if not create:
            sites = self.virtual_cluster._site[str(wp)]
            vc_basis = self.virtual_cluster._basis[str(wp)]
            return vc_basis, sites

        sites = self.wyckoff.position(wp, default=True)
        sites_c = to_cartesian(self.symmetry_operation.crystal, sites)

        basis = []
        info = []
        for rank in range(12):
            hs = self.harmonics.select(rank=rank, head="Q")
            for h in hs:
                fp = NSArray([h.expression(v=sites_c[p]).tolist() for p in range(len(sites_c))], "vector").expand()
                if np.any(fp != sp.S(0)):
                    basis.append(fp)
                    info.append(h.tag)

        basis = NSArray.concat(basis)

        if ortho:
            basis, idx = NSArray.orthogonalize(basis, len(sites))
            basis = basis[idx]
            info = [info[i] for i in idx]

        vc_basis = {i: basis[no].expand() for no, i in enumerate(info)}

        return vc_basis, sites

    # ==================================================
    def site_cluster_samb(self, site):
        """
        create site-cluster multipole basis set.

        Args:
            site (str or NSArray): representative site, "[x,y,z]".

        Returns: tuple.
            - dict: site-cluster SAMB, {TagMultipole: NSArray(vector)}.
            - NSArray: sites, NSArray(vector).
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        site_map = self.site_mapping(site)
        vc = self.virtual_cluster

        info = vc.key_list()
        bs = []
        n = len(site_map.keys())
        for basis in vc.values():
            s_basis = NSArray(str([0 for _ in range(n)]))
            for si, nos in enumerate(site_map.values()):
                for no in nos:
                    s_basis[si] += basis[no]
            bs.append(s_basis)

        bs = NSArray.concat(bs)
        bs, idx = NSArray.orthogonalize(bs)

        sc_samb = {info[i].replace(m_type="s"): bs[i].expand().simplify() for i in idx}
        site = NSArray.from_str(site_map.keys())

        return sc_samb, site

    # ==================================================
    def bond_cluster_samb(self, bond):
        """
        create bond-cluster multipole basis set.

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".

        Returns: tuple.
            - dict: bond-cluster SAMB, {TagMultipole: NSArray(vector)}.
            - NSArray: bonds, NSArray(bond).
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        bond_map, _ = self.bond_mapping(bond)
        vc = self.virtual_cluster

        info = vc.key_list()
        n = len(bond_map.keys())
        bs_s = []
        bs_a = []
        for basis in vc.values():
            bs_basis = NSArray(str([0 for _ in range(n)]))
            ba_basis = NSArray(str([0 for _ in range(n)]))
            for bo, nos in enumerate(bond_map.values()):
                for no in nos:
                    bs_basis[bo] += basis[abs(no)]
                    if no < 0:
                        ba_basis[bo] -= basis[abs(no)]
                    else:
                        ba_basis[bo] += basis[abs(no)]
            bs_s.append(bs_basis)
            bs_a.append(ba_basis)

        bs_s = NSArray.concat(bs_s)
        bs_a = NSArray.concat(bs_a)
        bs_s, idx_s = NSArray.orthogonalize(bs_s)
        bs_a, idx_a = NSArray.orthogonalize(bs_a)

        sg_basis_s = {info[i].replace(m_type="b"): bs_s[i].expand().simplify() for i in idx_s}
        sg_basis_a = {info[i].replace(m_type="b", head="T"): sp.I * bs_a[i].expand().simplify() for i in idx_a}

        bc_samb = sg_basis_s | sg_basis_a
        bond = NSArray.from_str(bond_map.keys())

        return bc_samb, bond

    # ==================================================
    def atomic_samb(self, bra_list, ket_list, spinful=False):
        """
        create atomic multipole basis set.

        Args:
            bra_list ([str]): atomic basis set for bra. see the examples in :class:`.AtomicBasisSet`
            ket_list ([str]): atomic basis set for ket. see the examples in :class:`.AtomicBasisSet`
            spinful (bool, optional): spinful ?

        Returns:
            dict: atomic SAMB, {TagMultipole: NSArray(matrix)}.
        """
        b_type = basis_type(bra_list[0], self.tag.crystal)

        bam = self.mp_dataset[b_type]
        hs = self.harmonics

        a_samb = create_atomic_samb(bra_list, ket_list, spinful, self.tag.crystal, bam, hs, ortho=True)

        return a_samb

    # ==================================================
    def z_samb(self, x_tag_list, y_tag_list, toroidal_priority=False, **kwargs):
        """
        create combined multipole basis set.

        Args:
            x_tag_list (str/TagMultipole/[TagMultipole]):  multipole/harmonics tag list.
            y_tag_list (str/TagMultipole/[TagMultipole]):  multipole/harmonics tag list.
            toroidal_priority (bool, optional): create toroidal multipoles (G,T) in priority? else prioritize conventional multipoles (Q,M).
            kwargs (dict, optional): select condition for multipoles, keywords in TagMultipole except for head.

        Returns:
            dict: {(TagMultipole, TagMultipole(atomic), TagMultipole(site/bond)): [(coefficient, TagMultipole(atomic), TagMultipole(site/bond)] }.
        """
        if type(x_tag_list) == str:
            x_tag_list = TagList([TagMultipole(x_tag_list)])
        if type(x_tag_list) in (TagMultipole, list):
            x_tag_list = TagList(x_tag_list)

        if type(y_tag_list) == str:
            y_tag_list = TagList([TagMultipole(y_tag_list)])
        if type(y_tag_list) in (TagMultipole, list):
            y_tag_list = TagList(y_tag_list)

        cg = self.clebsch_gordan
        hs = self.harmonics

        z_samb = create_z_samb(cg, hs, x_tag_list, y_tag_list, toroidal_priority, **kwargs)

        return z_samb

    # ==================================================
    @classmethod
    def spherical_atomic_multipole_basis(cls, bra_list, ket_list, spinful=False, crystal="cubic", core=None, verbose=False):
        """
        create spherical atomic multipole basis set.

        Args:
            bra_list ([str]): atomic basis set for bra. see the examples in :class:`.AtomicBasisSet`
            ket_list ([str]): atomic basis set for ket. see the examples in :class:`.AtomicBasisSet`
            spinful (bool, optional): spinful ?
            crystal (str, optional): seven crystal systems, triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic.
            core (BinaryManager, optional): core for binary data.
            verbose (bool, optional): verbose parallel info and verbose access to binary data ?

        Returns:
            dict: spherical atomic SAMB, {TagMultipole: NSArray(matrix)}.
        """
        if core is None:
            core = get_binary(verbose=verbose)

        b_type = basis_type(bra_list[0], crystal)

        bam = core[BaseAtomicMultipoleDataset][b_type]
        hs = None

        a_samb = create_atomic_samb(bra_list, ket_list, spinful, crystal, bam, hs, ortho=False)

        return a_samb



"""
TagGroup manages tag of group.
"""
from dataclasses import dataclass
from multipie.tag.tag import Tag
from multipie.tag.tag_list import TagList
from multipie.data.data_tag_space_group import _data_tag_space_group, _data_tag_space_group_crystal
from multipie.data.data_tag_point_group import (
    _data_tag_point_group,
    _data_tag_point_group_complex,
    _data_tag_point_group_real,
    _data_tag_point_group_crystal,
    _data_tag_point_group_cubic_series,
    _data_tag_point_group_hexagonal_series,
)
from multipie.const import __def_dict__


# ==================================================
@dataclass(frozen=True, order=True)
class TagGroup(Tag):
    """
    tag of group.
    """

    no: int = 1  # group No.
    SS: str = "C_1"  # Schoenflies symbol in LaTeX.
    IS: str = "C_1"  # international symbol(short) in LaTeX.
    crystal: str = "triclinic"  # crystal system.
    setting: str = ""  # setting.
    pg: str = ""  # point-group tag for space group, or "" for point group.
    subgroup: str = "cubic"  # subgroup (cubic/hexagonal).

    # ==================================================
    @classmethod
    def _from_str(cls, tag):
        if tag.count("^") > 0:
            no, SS, IS, setting, crystal, pg = _data_tag_space_group[tag]
        else:
            no, SS, IS, setting, crystal = _data_tag_point_group[tag]
            pg = ""

        subgroup = __def_dict__["subgroup"][crystal]

        d = {"no": no, "SS": SS, "IS": IS, "setting": setting, "crystal": crystal, "pg": pg, "subgroup": subgroup}

        return d

    # ==================================================
    def __str__(self):
        return self.SS.replace("_", "").replace("{", "").replace("}", "")

    # ==================================================
    def latex(self):
        return self.SS

    # ==================================================
    def info(self, latex=False):
        """
        info. of point group.

        Args:
            latex (bool, optional): in LaTeX ?

        Returns:
            - int: No. of group.
            - str: Schoenflies symbol.
            - str: International (short) symbol.
            - str: setting.
            or
            - str: info. string in LaTeX.
        """
        if latex:
            sp = r"\quad"
            r = f"No. {self.no}{sp}${self.SS}${sp}${self.IS}$"
            if self.setting:
                r += f"{sp}({self.setting} setting)"
            r += f"{sp}[ {self.crystal} ]"
            return r
        else:
            return self.no, self.SS, self.IS, self.setting

    # ==================================================
    def is_point_group(self):
        """
        is point group ?

        Returns:
            bool: is point group ?
        """
        return self.pg == ""

    # ==================================================
    def is_hexagonal_subgroup(self):
        """
        is hexagonal subgroup ?

        Returns:
            bool: is hexagonal subgroup ?
        """
        return self.subgroup == "hexagonal"

    # ==================================================
    @property
    def lattice(self):
        """
        Bravais lattice, (A/B/C/P/I/F/R) or null for point group.

        Returns:
            str: lattice letter.
        """
        if self.is_point_group():
            return ""
        else:
            return self.IS[0]

    # ==================================================
    @classmethod
    def create(cls, crystal="", space_group=False):
        """
        create group tags.

        Args:
            crystal (str, optional): crystal system, or "complex/real/cubic_series/hexagonal_series" for point group.
            space_group (bool, optional): space group or point_group ?

        Returns:
            TagList: tags.
        """
        if space_group:
            if crystal in __def_dict__["crystal"]:
                return TagList.from_str(TagGroup, _data_tag_space_group_crystal[crystal])
            else:
                return TagList.from_str(TagGroup, _data_tag_space_group.keys())
        else:
            if crystal in __def_dict__["crystal"]:
                return TagList.from_str(TagGroup, _data_tag_point_group_crystal[crystal])
            elif crystal == "complex":
                return TagList.from_str(TagGroup, _data_tag_point_group_complex)
            elif crystal == "real":
                return TagList.from_str(TagGroup, _data_tag_point_group_real)
            elif crystal == "cubic_series":
                return TagList.from_str(TagGroup, _data_tag_point_group_cubic_series)
            elif crystal == "hexagonal_series":
                return TagList.from_str(TagGroup, _data_tag_point_group_hexagonal_series)
            else:
                return TagList.from_str(TagGroup, _data_tag_point_group.keys())



from gcoreutils.convert_util import text_to_sympy
from gcoreutils.nsarray import NSArray


def convert_to_matrix(n_bra, n_ket, lst):
    mat = NSArray.zeros((n_bra, n_ket), "matrix")
    for i, j, m in lst:
        mat[i, j] = text_to_sympy(m)
    return mat



"""
SpaceGroup manages space group.
"""
import sympy as sp
import numpy as np
from gcoreutils.nsarray import NSArray
from multipie.data.data_no_space_group import _data_no_space_group
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_list import TagList
from multipie.tag.tag_multipole import TagMultipole
from multipie.group.point_group import PointGroup
from multipie.symmetry_operation.util.symmetry_operation_util import to_conventional, to_primitive
from multipie.symmetry_operation.symmetry_operation_g import SymmetryOperationG
from multipie.symmetry_operation.symmetry_operation_g_set import SymmetryOperationGSet
from multipie.wyckoff.wyckoff_g_set import WyckoffGSet
from multipie.wyckoff.wyckoff_g import WyckoffG
from multipie.multipole.util.z_samb_util import create_z_samb
from multipie import get_binary


# ==================================================
class SpaceGroup:
    """
    space group.

    Attributes:
        tag (TagGroup): space-group tag.
        symmetry_operation (SymmetryOperationG): the class to manage symmetry operations.
        wyckoff (WyckoffG): the class to manage Wyckoff positions.
        pg (PointGroup): associated point group.
        core (BinaryManager): core for binary data.
    """

    # ==================================================
    def __init__(self, sg_tag=None, core=None, verbose=False):
        """
        initialize the class.

        Args:
            sg_tag (TagGroup or str or int, optional): tag of space group.
            core (BinaryManager, optional): core for binary data.
            verbose (bool, optional): verbose access to binary data ?

        Notes:
            - if sg_tag is None, "C1^1" is used.
        """
        if sg_tag is None:
            sg_tag = "C1^1"
        if type(sg_tag) == int:
            sg_tag = _data_no_space_group[sg_tag]
        if type(sg_tag) == str:
            sg_tag = TagGroup(sg_tag)
        assert sg_tag in TagGroup.create(space_group=True), f"{sg_tag} is not a space-group tag."

        if core is None:
            core = get_binary(verbose=verbose)

        self.tag = sg_tag
        """space-group tag."""

        self.symmetry_operation: SymmetryOperationG = core[SymmetryOperationGSet][sg_tag]
        """the class to manage symmetry operations."""

        self.wyckoff: WyckoffG = core[WyckoffGSet][sg_tag]
        """the class to manage Wyckoff positions."""

        self.pg = PointGroup(self.tag.pg, core=core, verbose=verbose)
        """associated point group."""

        self.core = core
        """binary data."""

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def transform_matrix_site(self, site, cc_only=False):
        """
        transform matrix for site basis (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".
            cc_only (bool, optional): conjugacy-class operations only ?

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: site basis.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        s = self.symmetry_operation._equivalent_vector(site, primitive=True, shift=True, remove_duplicate=True)
        op = self.symmetry_operation.mat(cc_only=cc_only, primitive=True)

        tm = []
        for ii in range(len(op)):
            opi = op[ii]
            tmg = np.zeros((len(s), len(s))).astype(int).tolist()
            for i in range(len(s)):
                si = s[i].transform(opi).shift()
                tmg[s.index(si)][i] = 1
            tm.append(tmg)

        tm = NSArray(tm, "matrix")
        s = to_conventional(self.symmetry_operation.lattice, s).shift()

        return tm, s

    # ==================================================
    def transform_matrix_bond(self, bond, nondirectional=False, cc_only=False):
        """
        transform matrix for bond basis (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            nondirectional (bool, optional): ignore directional property ?
            cc_only (bool, optional): conjugacy-class operations only ?

        Returns: tuple
            - NSArray: symmetry operation matrices.
            - NSArray: bond basis.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        b = self.symmetry_operation._equivalent_bond(
            bond, primitive=True, nondirectional=nondirectional, shift=True, remove_duplicate=True
        )
        op = self.symmetry_operation.mat(cc_only=cc_only, primitive=True)

        tm = []
        for ii in range(len(op)):
            opi = op[ii]
            tmg = np.zeros((len(b), len(b))).astype(int).tolist()
            for i in range(len(b)):
                bi = b[i].transform(opi).shift()
                if nondirectional:
                    j = b.index(bi)
                    if j is None:
                        j = b.index(bi.reverse_direction())
                else:
                    j = b.index(bi)
                tmg[j][i] = 1
            tm.append(tmg)

        tm = NSArray(tm, "matrix")
        b = to_conventional(self.symmetry_operation.lattice, b).shift()

        return tm, b

    # ==================================================
    def transform_site(self, site, shift=False, remove_duplicate=False, plus_set=False):
        """
        set of transformed sites (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".
            shift (bool, optional): transformed sites are moved into unit cell ?
            remove_duplicate (bool, optional): remove the same sites ?
            plus_set (bool, optional): add partial translations ?

        Returns:
            NSArray: transformed sites (in order of symmetry operations, and plus_set).

        Notes:
            - when remove_duplicate and shift are True, return values in the same order as site_mapping keys.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        if remove_duplicate and shift:
            s = NSArray.from_str(self.site_mapping(site, plus_set).keys())
        else:
            s = self.symmetry_operation._equivalent_vector(
                site, plus_set=plus_set, shift=shift, remove_duplicate=remove_duplicate
            )

        return s

    # ==================================================
    def transform_bond(self, bond, nondirectional=False, shift=False, remove_duplicate=False, plus_set=False):
        """
        set of transformed bonds (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            nondirectional (bool, optional): ignore directional property ?
            shift (bool, optional): transformed bond starting points are moved into unit cell ?
            remove_duplicate (bool, optional): remove the same bonds ?
            plus_set (bool, optional): add partial translations ?

        Returns:
            NSArray: transformed bonds (in order of symmetry operations, and plus_set).

        Notes:
            - when remove_duplicate, nondirectional and shift are True, return values in the same order as bond_mapping keys.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        if remove_duplicate and nondirectional and shift:
            b = NSArray.from_str(self.bond_mapping(bond, plus_set=plus_set)[0].keys())
        else:
            b = self.symmetry_operation._equivalent_bond(
                bond, plus_set=plus_set, nondirectional=nondirectional, shift=shift, remove_duplicate=remove_duplicate
            )

        return b

    # ==================================================
    def shift_home_unit_cell(self, site):
        """
        shift to home unit cell.

        Args:
            site (str or NSArray): site to be shifted.

        Returns:
            NSArray: shifted site.
        """
        if type(site) == str:
            site = NSArray(site)
        lattice = self.symmetry_operation.lattice
        site = to_conventional(lattice, to_primitive(lattice, site).shift()).shift()

        return site

    # ==================================================
    def find_rep_site(self, site_list):
        """
        find representative sites from site_list.

        Args:
            site_list (NSArray): site list.

        Returns:
            NSArray: representative sites.
        """

        def remove_equivalent_site(sl, rep_site):
            if not sl:
                return rep_site
            else:
                s = list(sl)[0]
                r = set(self.transform_site(s, shift=True, remove_duplicate=True).str())
                rep_site.append(s)
                return remove_equivalent_site(sl - r, rep_site)

        assert len(site_list) > 0, "empty site list is given."

        lattice = self.symmetry_operation.lattice
        site_list = set(to_conventional(lattice, to_primitive(lattice, site_list).shift().remove_duplicate()).shift().str())

        rep_site = []
        rep_site = remove_equivalent_site(site_list, rep_site)
        rep_site = NSArray.from_str(rep_site)

        return rep_site

    # ==================================================
    def find_rep_bond(self, bond_list):
        """
        find representative bonds from bond_list.

        Args:
            bond_list (NSArray): bond list.

        Returns:
            NSArray: representative bonds.
        """

        def remove_equivalent_bond(bl, rep_bond):
            if not bl:
                return rep_bond
            else:
                b = list(bl)[0]
                r = self.transform_bond(b, shift=True, remove_duplicate=True)
                r = NSArray.concat([r, r.reverse_direction()])
                r = set(r.str())
                rep_bond.append(b)
                return remove_equivalent_bond(bl - r, rep_bond)

        assert len(bond_list) > 0, "empty bond list is given."

        # remove equivalent bonds.
        tail = bond_list[0].convert_bond("bond_th")[0]
        lattice = self.symmetry_operation.lattice
        bond_list = set(to_conventional(lattice, to_primitive(lattice, bond_list).shift().remove_duplicate()).shift().str())

        # find representative bonds.
        rep_bond = []
        rep_bond = remove_equivalent_bond(bond_list, rep_bond)
        rep_bond = NSArray.from_str(rep_bond)

        # set bonds with given tail.
        tail = to_primitive(lattice, tail).shift()
        rep_bond1 = []
        for bond in rep_bond:
            bonds = self.symmetry_operation._equivalent_bond(bond, remove_duplicate=True, primitive=True, shift=True)
            t, h = bonds.convert_bond("bond_th")
            idx = t.shift().index(tail)
            if idx is None:
                idx = h.shift().index(tail)
                assert idx is not None, "not reach here."
                b = bonds[idx].reverse_direction()
            else:
                b = bonds[idx]
            v, c = b.shift().convert_bond("bond")
            b = NSArray(f"{v}@{c}")
            b = to_conventional(lattice, b)
            rep_bond1.append(str(b))
        rep_bond = NSArray.from_str(rep_bond1)

        return rep_bond

    # ==================================================
    def site_mapping(self, site, plus_set=False):
        """
        mapping between site and symmetry operations (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".
            plus_set (bool, optional): add partial translations ?

        Returns:
            dict: mapping from site string to symmetry-operation IDs.

        Notes:
            - sites are sorted by 1st SO.
            - if plus_set is True, return [set(t0),set(t1),...]
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for vector/site.")

        s = self.symmetry_operation._equivalent_vector(site, primitive=True, shift=True)
        so_num = len(self.symmetry_operation.full)

        lst = {}
        for i in range(len(s)):
            si = str(s[i])
            lst[si] = lst.get(si, []) + [i % so_num]

        # sort in order of 1st component of SO.
        lst = dict(sorted(lst.items(), key=lambda i: i[1][0]))

        # convert to conventional cell.
        s = NSArray.from_str(lst.keys())
        s = to_conventional(self.symmetry_operation.lattice, s, plus_set=plus_set).shift()

        if plus_set:
            mpp = list(lst.values()) * len(self.symmetry_operation.plus_set)
            lst = {str(s[i]): mpp[i] for i in range(len(s))}

        return lst

    # ==================================================
    def bond_mapping(self, bond, plus_set=False):
        """
        mapping between bond and symmetry operations (reduced coordinate).

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            plus_set (bool, optional): add partial translations ?

        Returns: tuple
            - dict: mapping from bond to symmetry-operations IDs. bond = "tail;head"
            - bool: nondirectional ?

        Notes:
            - if bond direction is reversed, symmetry operation ID with minus sign.
            - bonds are sorted by 1st SO.
            - if plus_set is True, return [set(t0),set(t1),...]
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")
        if type(bond) == str:
            bond = NSArray(bond)

        b_d = self.symmetry_operation._equivalent_bond(
            bond, primitive=True, shift=True, nondirectional=False, remove_duplicate=True
        )
        b_all = self.symmetry_operation._equivalent_bond(
            bond, primitive=True, shift=True, nondirectional=False, remove_duplicate=False
        )
        b_nd = self.symmetry_operation._equivalent_bond(
            bond, primitive=True, shift=True, nondirectional=True, remove_duplicate=True
        )

        nd = len(b_nd) != len(b_d)
        b = b_nd if nd else b_d

        tail0 = bond.convert_bond("bond_th")[0]
        tail0 = NSArray.from_str(self.site_mapping(tail0).keys())
        tail0 = to_primitive(self.symmetry_operation.lattice, tail0).shift()
        head0 = bond.convert_bond("bond_th")[1]
        head0 = NSArray.from_str(self.site_mapping(head0).keys())
        head0 = to_primitive(self.symmetry_operation.lattice, head0).shift()

        # set bond direction with the same tail as given one.
        if str(tail0.sort()) == str(head0.sort()):
            for i in range(len(b)):
                t, h = b[i].convert_bond("bond_th")
                t_idx = tail0.index(t.shift())
                h_idx = tail0.index(h.shift())
                if t_idx > h_idx:
                    b[i] = b[i].reverse_direction()

        # create mapping in primitive cell.
        lst1 = {}
        for i in range(len(b_all)):
            bi = b_all[i]
            bib = str(bi.reverse_direction())
            bi = str(bi)
            if b.index(bi) is not None:
                lst1[bi] = lst1.get(bi, []) + [i + 1]
            elif b.index(bib) is not None:
                lst1[bib] = lst1.get(bib, []) + [-(i + 1)]
            else:
                assert False, "not reach here."

        # reverse SO if identity SO idx are negative.
        if sum(list(lst1.values()), []).count(-1) > 0:
            lst = {i: [-k for k in j] for i, j in lst1.items()}
        else:
            lst = lst1

        # reset SO number staring from zero.
        lsts = {}
        for i, j in lst.items():
            lsts[i] = [k - 1 if k > 0 else k + 1 for k in j]

        # sort in order of 1st component of SO.
        lsts = dict(sorted(lsts.items(), key=lambda i: abs(i[1][0])))

        # convert to conventional cell.
        b = NSArray.from_str(lsts.keys())
        b = to_conventional(self.symmetry_operation.lattice, b, plus_set=plus_set).shift()

        if plus_set:
            mp = list(lsts.values()) * len(self.symmetry_operation.plus_set)
            lsts = {str(b[i]): mp[i] for i in range(len(b))}

        return lsts, nd

    # ==================================================
    def find_wyckoff_position(self, site):
        """
        find Wyckoff position for a given site (reduced coordinate).

        Args:
            site (str or NSArray): site, "[x,y,z]".

        Returns:
            str: Wyckoff position.
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        wpv = self.symmetry_operation._equivalent_vector(site, plus_set=True, shift=True, remove_duplicate=True)
        n = len(wpv)
        tags = [i for i in self.wyckoff.keys() if i.n == n]
        if len(tags) == 1:
            return tags[0]

        wps = [self.wyckoff.position(tag) for tag in tags]
        wpv0 = self.symmetry_operation._equivalent_vector(site, primitive=True, shift=True, remove_duplicate=True)
        w0 = to_conventional(self.symmetry_operation.lattice, wpv0[0:1])[0]
        sht = [[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1]]
        for tag, wp in zip(tags, wps):
            for i in wp:
                for s in sht:
                    s = NSArray(s, "vector")
                    if str(i) == str(w0 + s):
                        return tag
                    if sp.solve(list(i - w0 + s), self.wyckoff.v.tolist(), manual=True):
                        return tag

        raise Exception(f"{str(site)} cannot be found.")

    # ==================================================
    def site_cluster_samb(self, site):
        """
        create site-cluster multipole basis set.

        Args:
            site (str or NSArray): representative site, "[x,y,z]".

        Returns: tuple,
            - dict: site-cluster SAMB, {TagMultipole: NSArray(vector)}.
            - NSArray: sites, NSArray(vector).
        """
        if not isinstance(site, (str, NSArray)):
            raise KeyError(f"{type(site)} is not accepted for site.")

        site_map = self.site_mapping(site)
        vc = self.pg.virtual_cluster

        info = vc.key_list()
        bs = []
        n = len(site_map.keys())
        for basis in vc.values():
            s_basis = NSArray(str([0 for _ in range(n)]))
            for si, nos in enumerate(site_map.values()):
                for no in nos:
                    s_basis[si] += basis[no]
            bs.append(s_basis)

        bs = NSArray.concat(bs)
        bs, idx = NSArray.orthogonalize(bs)

        sc_samb = {info[i].replace(m_type="s"): bs[i].expand().simplify() for i in idx}
        site = NSArray.from_str(site_map.keys())

        return sc_samb, site

    # ==================================================
    def bond_cluster_samb(self, bond):
        """
        create bond-cluster multipole basis set.

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".

        Returns: tuple,
            - dict: bond-cluster SAMB, {TagMultipole: NSArray(vector)}.
            - NSArray: bonds, NSArray(bond).
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        bond_map, _ = self.bond_mapping(bond)
        vc = self.pg.virtual_cluster

        info = vc.key_list()
        n = len(bond_map.keys())
        bs_s = []
        bs_a = []
        for basis in vc.values():
            bs_basis = NSArray(str([0 for _ in range(n)]))
            ba_basis = NSArray(str([0 for _ in range(n)]))
            for bo, nos in enumerate(bond_map.values()):
                for no in nos:
                    bs_basis[bo] += basis[abs(no)]
                    if no < 0:
                        ba_basis[bo] -= basis[abs(no)]
                    else:
                        ba_basis[bo] += basis[abs(no)]
            bs_s.append(bs_basis)
            bs_a.append(ba_basis)

        bs_s = NSArray.concat(bs_s)
        bs_a = NSArray.concat(bs_a)
        bs_s, idx_s = NSArray.orthogonalize(bs_s)
        bs_a, idx_a = NSArray.orthogonalize(bs_a)

        sg_basis_s = {info[i].replace(m_type="b"): bs_s[i].expand().simplify() for i in idx_s}
        sg_basis_a = {info[i].replace(m_type="b", head="T"): sp.I * bs_a[i].expand().simplify() for i in idx_a}

        bc_samb = sg_basis_s | sg_basis_a
        bond = NSArray.from_str(bond_map.keys())

        return bc_samb, bond

    # ==================================================
    def z_samb(self, x_tag_list, y_tag_list, toroidal_priority=False, **kwargs):
        """
        create combined multipole basis set.

        Args:
            x_tag_list (str/TagMultipole/[TagMultipole]):  multipole/harmonics tag list.
            y_tag_list (str/TagMultipole/[TagMultipole]):  multipole/harmonics tag list.
            toroidal_priority (bool, optional): create toroidal multipoles (G,T) in priority? else prioritize conventional multipoles (Q,M).
            kwargs (dict, optional): select condition for multipoles, keywords in TagMultipole except for head.

        Returns:
            dict: {(TagMultipole, TagMultipole(atomic), TagMultipole(site/bond)): [(coefficient, TagMultipole(atomic), TagMultipole(site/bond)] }.
        """
        if type(x_tag_list) == str:
            x_tag_list = TagList([TagMultipole(x_tag_list)])
        if type(x_tag_list) in (TagMultipole, list):
            x_tag_list = TagList(x_tag_list)

        if type(y_tag_list) == str:
            y_tag_list = TagList([TagMultipole(y_tag_list)])
        if type(y_tag_list) in (TagMultipole, list):
            y_tag_list = TagList(y_tag_list)

        cg = self.pg.clebsch_gordan
        hs = self.pg.harmonics

        z_samb = create_z_samb(cg, hs, x_tag_list, y_tag_list, toroidal_priority, **kwargs)

        return z_samb
